package Impedance::Complex;
#$Id: Complex.pm,v 1.33 2016/02/03 13:53:09 sofc Exp $
###########################################################
#
#
# Package Impedance::Complex;
#
# This module defines series and parallel connections of
# impedance elements. The resulting element can be treated
# as any other Impedance::Base derived element, thus it is
# possible using the serial and parralel elements to build
# arbitrary complex circuit layouts!
#
# In effect the module defines two container types which from 
# the outside behaves as a single impedance element. The 
# Impecance::Complex container elements together with the 
# Impedance::Base derived elements implements the Composite pattern
# with the container elements (Impedance::Complex derived) 
# acting as GoF Composite classes
# and the Impedance::Base derived classes acting as Gof Leaf classes
#
# To obtain an element instance, call one of the constructors:
#
# $id = Impedance::Ser->new($id);
#
# $id = Impedance::Par->new($id);
#
# The module defines four new member functions:
#
# $id->elements()      Returns a list of Impedance::Base (or 
#                      derived) elements. 
# $id->elements_all()  Returns the complete element list for all
#                      elements in the container (recursively).
# $id->element($)      Returns the element instance with the 
#                      specified id (usually obtained from the
#                      elements member function).
# $id->add_element($$) Adds an element or the list of elements
#                      contained within the element instance.
#                      The arguments are type and id of the
#                      element to add. It returns the element
#                      instance created. The element is created
#                      using the Impedance::Device::new_device($$)
#                      factory function.
#
# The print_line() function inherited from Impedance::Base
# is overloaded with slightly different behaviour.
# Instead of a string containing just a single line, the return
# value of print_line() contains multiple lines. The first
# line contains a start identifier (either '(' or '[' 
# depending on type) and the last line contains the corresponding
# end identifier ( ')' or ']' ).
# All intermediate lines are obtained by calling print_line()
# on the individual element(s) in the container.
# Note that this may include additional complex elements
# resulting in nested parantheses (Which is completely valid
# behaviour)!
#
# Additionally the tags() function is also overloaded so that
# it returns a list of the elements in the container instead 
# of valid parameter names (as the complex element is a container
# it does not by itself have any parameters of which to set or 
# get the value).
#
# It is also not possible to define any help functions on complex 
# elements, thus any calls to helperfunctions() merely results 
# in the empty list.
#
# Notice that f_imax() can only report summit frequencies between 
# 1e-100 and 1E100 Hz Thus if the true sumit
# frequency of the complex element used is outside this range it 
# will be reported wrongly!
# Additionally, the f_imax() function first tries to determine a
# summit frequency in the R,-X quadrant (corresponding to where 
# an arc from a parallewl RC circuit will be found). If no summit 
# frequency is found it reports the summit frequency in the R,X 
# plane (Where an parallel R-L circuit will be found).
#
# In addition to the container elements described above the 
# IMpedance::Complex module also defines the following compound 
# container elements:
#
# $id = Impedance::TLShort->new($id);
#
# $id = Impedance::TLBlock->new($id);
#
# Unlike the Ser and Par elements these container elements have 
# intrinsic parameters in addition to the parameters from the 
# elements conteined inside them and thus they behave both as 
# simple elements and as container elements.
#
# The TLShort and TLBlock elements model transmission lines
# where the branch to branch impedance can be customized (as opposed 
# to the normal warburg type elements where the branch to branch 
# impedance is assumed to be a capacitor).
#
###########################################################
use strict;
use warnings;
use Impedance::Base;
use Math::Complex;
our @ISA = qw(Impedance::Base);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'Complex';
    $self->{'_tags'} = [];
    $self->{'_helptags'} = [];
    $self->{'_devices'} = [];
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    my $imax = 0;
    my $imin = 0;
    my $fimax;
    my $fimin;
    for (my $x = -100; $x < 100; $x+= 0.2)
    {
	my $f = 10**$x;
	my $im = -Im($self->F($f));
	if ($imax < $im)
	{
	    $imax = $im;
	    $fimax = $f;	    
	}
	if ($imin > $im)
	{
	    $imin = $im;
	    $fimin = $f;	    
	}
    }
    return $fimax if ($fimax);
    return $fimin;
}

sub fit()
{
    my $self = shift;
    my @l = ();
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->fit();
    }
    return @l;
}

sub elements()
{
    my $self = shift;
    return @{$self->{'_devices'}}
}

sub elements_all()
{
    my $self = shift;
    my @devlist = ();
    foreach my $dev (@{$self->{'_devices'}})
    {
	push @devlist,$dev;
	if (ref($dev) =~ /Par|Ser|[a-zA-Z]+\_$/)
	{
	    foreach ($dev->elements_all)
	    {
		push @devlist,$_;
	    }
	}
    }
    return @devlist;
}


sub element($)
{
    my $self = shift;
    my $e = shift;
    foreach (@{$self->{'_devices'}})
    {
	return $_ if ($_ eq $e);
    }
}

sub value()
{
    my $self = shift;
    return undef;
}

sub tags()
{
    my $self = shift;
    return @{$self->{'_devices'}};
}

sub save()
{
    my $self = shift;
    my @l = ();
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->save;
    }
    return join("\n",@l),"\n";
}

sub add_element($$)
{
    my $self = shift;
    my $type = shift;
    my $nr = shift;
    my $id = Impedance::Device::new_device($type,$nr);
    push @{$self->{'_devices'}},$id;
    return $id;
}

package Impedance::Par;
use Math::Complex;
our @ISA = qw(Impedance::Complex);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'Parallel';
    $self->{'_tags'} = [];
    $self->{'_helptags'} = [];
    bless $self,$class;
    return $self;
}

sub print_line()
{
    my $self = shift;
    my @l = ();
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->print_line();
    }
    return join("\n",'(',@l,')');
}

sub name()
{
    my $self = shift;
    my @l = ();
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->name;
    }
    return '('.join(' ',@l).')';
}

sub functions()
{
    my $self = shift;
    my @l = ();
    my @l2 = ();
    my $n = $self->{'_nr'};
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->functions;
	push @l2,'1.0/'.$d->function_name;
    }
    push @l,'z_'.$n.'(x) = 1.0/( '.join(' + ',@l2).' )';
    return join("\n",@l),"\n";
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    my $res = cplx(0,0);
    foreach my $d (@{$self->{'_devices'}})
    {
	$res += 1.0/$d->F($f);
    }
    return 1.0/$res;
}

package Impedance::Ser;
use Math::Complex;
our @ISA = qw(Impedance::Complex);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'Serial';
    $self->{'_tags'} = [];
    $self->{'_helptags'} = [];
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    my $imax = 0;
    my $fimax;
    foreach my $d (@{$self->{'_devices'}})
    {
	my $f = $d->f_imax();
	my $im = $d->F($f);
	next unless $imax < $im;
	$imax = $im;
	$fimax = $f;
    }
    return $fimax;
}

sub print_line()
{
    my $self = shift;
    my @l = ();
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->print_line();
    }
    return join("\n",'[',@l,']');
}

sub name()
{
    my $self = shift;
    my @l = ();
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->name;
    }
    return '['.join(' ',@l).']';
}

sub functions()
{
    my $self = shift;
    my @l = ();
    my @l2 = ();
    my $n = $self->{'_nr'};
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->functions;
	push @l2,$d->function_name;
    }
    push @l,'z_'.$n.'(x) = '.join(' + ',@l2);
    return join("\n",@l),"\n";
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    my $res = cplx(0,0);
    foreach my $d (@{$self->{'_devices'}})
    {
	$res += $d->F($f);
    }
    return $res;
}


package Impedance::Container_;
use Math::Complex;
our @ISA = qw(Impedance::Complex);

### Base class for complex container elements which in addition to containing ###
### elements also have intrinsic variables                                    ###
### All classes deriving from this must end with an underscore for proper     ###
### parsing by Impedance::Model!                                              ###

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'Container';
    $self->{'_tags'} = [];
    $self->{'_helptags'} = [];
    $self->{'_desc'}{'name'} = 'Default container with tags';
    bless $self,$class;
    return $self;
}


sub print_line()
{
    my $self = shift;
    my @l = ();
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->print_line();
    }
    my $res = $self->type().' ';
    my @l2 = ();
    foreach my $t ($self->tags)
    {
        push @l2,$t.'='.$self->value($t);
    }
    return $res.join(', ',@l2)." (\n".join("\n",@l,')');
}

sub name()
{
    my $self = shift;
    my @l = ();
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->name();
    }
    return $self->{'_name'}.'('.join(' ',@l).')'.'_'.$self->{'_nr'};
}

sub value
{
    my $self = shift;
    return Impedance::Base::value($self,@_);
}

sub tags()
{
    my $self = shift;
    return (@{$self->{'_tags'}});
}

sub save()
{
    my $self = shift;
    my @l = ();
    my $n = $self->{'_nr'};
    foreach my $t (@{$self->{'_tags'}})
    {
        push @l,"Final_".$t."_$n = ".$t."_$n * f".lc($t)."_$n";
    }
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->save;
    }
    return join("\n",@l),"\n";
}

sub my_tanh($)
{
    my $self = shift;
    my $arg = shift;
    if ($arg > 100)
    {
        return cplx(1,0);
    }
    if ($arg < 1e-10)
    {
        return $arg;
    }
    else
    {
        return tanh($arg);
    }
}


sub my_coth($)
{
    my $self = shift;
    my $val = shift;
    if ($val > 50)
    {
        return cplx(1,0);
    }
    else
    {
        return cosh($val)/sinh($val);
    }
}


sub type()
{
    my $self = shift;
    return $self->{'_name'}.':';
}


package Impedance::RSerial_;
use Math::Complex;
our @ISA = qw(Impedance::Container_);
### Dummy device for testing Container_ base class ###

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'RSerial';
    $self->{'_tags'} = ['R'];
    $self->{'_helptags'} = [];
    $self->{'_desc'}{'name'} = 'Resistor in series';
    $self->{'_desc'}{'R'} = 'Resistance in Ohm';
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    my $imax = 0;
    my $fimax;
    foreach my $d (@{$self->{'_devices'}})
    {
	my $f = $d->f_imax();
	my $im = $d->F($f);
	next unless $imax < $im;
	$imax = $im;
	$fimax = $f;
    }
    return $fimax;
}

sub functions()
{
    my $self = shift;
    my @l2 = ();
    my $n = $self->{'_nr'};
    my @l = ("R_$n = ".$self->value('R'),"fr_$n = 1.0");
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->functions;
	push @l2,$d->function_name;
    }
    if (@l2)
    {
	push @l,'z_'.$n.'(x) = R_'.$n.'*fr_'.$n.' + '.join(' + ',@l2);
    }
    else
    {
	push @l,'z_'.$n.'(x) = R_'.$n.'*fr_'.$n;
    }
    return join("\n",@l),"\n";
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    my $res = cplx($self->value('R'),0);
    foreach my $d (@{$self->{'_devices'}})
    {
	$res += $d->F($f);
    }
    return $res;
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("fr_$n");
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->fit();
    }
    return @l;
}



package Impedance::TLS_;
use Math::Complex;
our @ISA = qw(Impedance::Container_);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'TLShort';
    $self->{'_tags'} = ['R','n'];
    $self->{'_values'}{'n'} = 0.5;
    $self->{'_fit_nodefault'} = {'n' => 1};
    $self->{'_helptags'} = [];
    $self->{'_desc'}{'name'} = 'Transmission line terminated with a short circuit';
    $self->{'_desc'}{'R'} = 'Line resistance in Ohm';
    $self->{'_desc'}{'n'} = 'Exponent';
    bless $self,$class;
    return $self;
}

sub functions()
{
    my $self = shift;
    my @l2 = ();
    my $n = $self->{'_nr'};
    my @l = ("R_$n = ".$self->value('R'),"fr_$n = 1.0","n_$n = ".$self->value('n'),"fn_$n = 1.0");
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->functions;
	push @l2,$d->function_name;
    }
    if (@l2)
    {
	push @l,'tnh(x) = (abs(x)>100 ? {1,0} : tanh(x))';
	push @l,'zx_'.$n.'(x) = '.join(' + ',@l2);
	push @l,'zr_'.$n.'(x) = R_'.$n.'*fr_'.$n;
	my $st = '(zr_'.$n.'(x)/zx_'.$n.'(x))**(n_'.$n.' * fn_'.$n.')';
	push @l,'z_'.$n.'(x) = zr_'.$n.'(x) * tnh('.$st.') / ('.$st.')';
    }
    else
    {
	push @l,'z_'.$n.'(x) = R_'.$n.'*fr_'.$n;
    }
    return join("\n",@l),"\n";
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    my $res = cplx($self->value('R'),0);
    return $res unless (@{$self->{'_devices'}});
    my $r_cont = cplx(0,0);
    foreach my $d (@{$self->{'_devices'}})
    {
	$r_cont += $d->F($f);
    }
    my $x = $res / $r_cont;
    $res = $res * $self->my_tanh($x**$self->value('n')) / $x**$self->value('n');
    return $res;
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("fr_$n","fn_$n");
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->fit();
    }
    return @l;
}


package Impedance::TLB_;
use Math::Complex;
our @ISA = qw(Impedance::Container_);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'TLBlock';
    $self->{'_tags'} = ['R'];
    $self->{'_helptags'} = [];
    $self->{'_values'}{'n'} = 0.5;
    $self->{'_fit_nodefault'} = {'n' => 1};
    $self->{'_desc'}{'name'} = 'Blocked transmission line';
    $self->{'_desc'}{'R'} = 'Line resistance in Ohm';
    $self->{'_desc'}{'n'} = 'Exponent';
    bless $self,$class;
    return $self;
}

sub functions()
{
    my $self = shift;
    my @l2 = ();
    my $n = $self->{'_nr'};
    my @l = ("R_$n = ".$self->value('R'),"fr_$n = 1.0","n_$n = ".$self->value('n'),"fn_$n = 1.0");
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->functions;
	push @l2,$d->function_name;
    }
    if (@l2)
    {
	push @l,'cth(x) = (abs(x)>50 ? {1,0} : cosh(x)/sinh(x))';
	push @l,'zx_'.$n.'(x) = '.join(' + ',@l2);
	push @l,'zr_'.$n.'(x) = R_'.$n.'*fr_'.$n;
	my $st = '(zr_'.$n.'(x)/zx_'.$n.'(x))**(n_'.$n.' * fn_'.$n.')';
	push @l,'z_'.$n.'(x) = zr_'.$n.'(x) * cth('.$st.')/('.$st.')';
    }
    else
    {
	push @l,'z_'.$n.'(x) = R_'.$n.'*fr_'.$n;
    }
    return join("\n",@l),"\n";
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    my $res = cplx($self->value('R'),0);
    return $res unless (@{$self->{'_devices'}});
    my $r_cont = cplx(0,0);
    foreach my $d (@{$self->{'_devices'}})
    {
	$r_cont += $d->F($f);
    }
    my $x = $res / $r_cont;
    $res = $res * $self->my_coth($x**$self->value('n')) / $x**$self->value('n');
    return $res;
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("fr_$n");
    foreach my $d (@{$self->{'_devices'}})
    {
	push @l,$d->fit();
    }
    return @l;
}

return 1;
