package Impedance::Model;
#$Id: Model.pm,v 1.88 2016/02/10 08:56:54 sofc Exp $
##############################################################
#
#  Package Impedance::Model
#
# This module defines how to handle impedance models. It utilises
# the impedance elements defined by Impedance::Base and the 
# derived classes.
#
# To obtain an instance, call one of the constructors:
#
# $id = Impedance::Model->new();
#
# $id = Impedance::Model->new($data);
#
# $id = Impedance::Model->new($model_ref);
#
# $id = Impedance::Model->new($model_ref,$minf,$maxf);
#
# The first constructor merely initialises a new empty model.
# The second constructor initalises an empty model but adds the
# impedance data in in the supplied data string to the model instance.
# The third constructor copies the model from the supplied model 
# reference and copies the data from that instance as well.
# The fourth constructor copies as the third, but only
# copies those data which lies within the frequency range specified
# by the last two arguments!
# This can be used for partial fitting where only a specific frequency
# range is needed.
#
# The individual impedance elements are strored in internal data structure
# which makes sure that only unique impedance element id's are used.
#
# The Impedance::Model class implements the Composite patern (together with
# the Impedance::Base and Impedance::Complex derived classes) with 
# Impedance::Model acting as a Gof Component class as well as a Gof Composite class.
#
# All Impedance::Model instance has the following public member functions:
#
# $id->parse($)              This function accepts a string defining an impedance
#                            model. The string must be lines of the format defined
#                            by the Impedance::Base function print_line().
#                            The function parses this string and initialises the
#                            correct impedance elements based on this.
#                            The function returns the id number of the last 
#                            impedance element added.
# $id->print_model()         Returns a string which can be parsed by $id->parse($)   
# $id->device_types()        Wrapper for Impedance::Device::list_device_types().
# $id->elements()            Returns a list of impedance element names (nr).
# $id->element($)            Returns the impedance element specified.
#                            The function also gets elements from inside containers!
# $id->elements_all()        Returns the complete element list of the model.
#                            Recursively calls into any container elements.
#                            Note that unlike the elements() function, the actual
#                            device instances are returned as opposed to the element 
#                            names.
# $id->model_text()          Returns a string representing the impedance model
# $id->delete_element($)     Deletes the specified impedance element.
# $id->add_element($[$])     Adds an impedance element of the specified type.
#                            If an additional arguemnt is specified, the id of
#                            the new element will be set to this number. If no 
#                            second argument is specified, hte next id is simply 
#                            chosen. The function returns a reference to the added
#                            element.
# $id->fit()                 Returns a list of possible fit paramerters for current model.  
# $id->data([$])             Gets or sets the impedance data in the internal data field.
# $id->delete_point($)       Deletes the point specified from the internal data array.
#                            Returns the new data in the same way as data().
#                            Note that index 0 is the point at the highest frequency.
# $id->print_fit([$])        Returns a string containing all the functions nescesarry 
#                            for gnuplot to initialise and fit the given model to the
#                            impedance data either in the specified filename or (in 
#                            absence of an argument) in the internal data field. All 
#                            data points are given even weight. 
# $id->print_fit_weight([$]) Same as print_fit(), however the actual fitting is done with
#                            uneven weight, so that data points with small values (absolute
#                            length of impedance vector) gets higher weight than points
#                            with larger values (weight inversely proportional to length).
# $id->set_show_arcs($)      Sets wether the plot functions should include the individual
#                            arcs in the plots or not. An argument of 0 disabels the arcs,
#                            1 includes them.
# $id->print_plot([$])       Returns a string containing the function declerations and 
#                            plot definitions for gnuplot to plot the specified data
#                            (if any) if no datafile name is specified, it uses the 
#                            data in the internal data field.
# $id->print_plot_range($$[$]) Similar to print_plot, except it must have the min and max
#                            range specified as the first 2 arguments.
#                            The effect is that any labels are only printed if they have
#                            x-values within the specified range.
# $id->print_bode([$])       Same as above, but for bode plots.
# $id->print_plot_eps([$])   Same as print_plot, but for eps file output. 
# $id->print_bode_eps([$])   Same as above, but for bode plots.
# $id->print_plot_error([$]) Returns a string containing the function declerations and 
#                            plot definitions for gnuplot to plot the difference between 
#                            the model and the specified data (if any) if no 
#                            datafile name is specified, it uses the 
#                            data in the internal data field.
# $id->print_imp_sim($$)     Returns the function declaratioins nescesarry for gnuplot
#                            to plot an impedance plot of the current model.
#                            Arguments are the frequency range to plot (min and max).
# $id->print_bode_sim($$)    Same as above, but for bode plot.
# $id->F($)                  Returns the impedance of the current model for the 
#                            specified frequency. The returned impedance is of type 
#                            Math::Complex.
# $id->subset($$)            Returns a subset of the data in the internal data field
#                            based on the specified minimum and maximum frequency.
# $id->subtract_model()      Returns the residual of the data in the data field once
#                            the impedance of the current model has been subtracted
#                            (subtraction done in the indivudual data points!).
# $id->get_error()           Returns the mean and maximum error for the data and model chosen
#                            The error is calculated as the absolute difference of the data 
#                            from the mode both the real and for the imaginary part. The error
#                            is then normalised with the modulus of the data value. 
#                            This is calculated for all frequencies in the data set and the
#                            mean and maximum values are returned (as percentages). 
# $id->minf()                Returns the minimum frequency in the data set.
# $id->maxf()                Returns the maximum frequency in the data set.
# $id->minr()                Returns the minimum real part of the impedance in the data set.
# $id->maxr()                Returns the maximum real part of the impedance in the data set.
# $id->mini()                Returns the minimum imaginary part of the impedance in the data set.
# $id->maxi()                Returns the maximum imaginary part of the impedance in the data set.
# $id->scale_factor([$])     Sets the scale factor function to the specified value.
#                            If no arguments, the scale factor function is set to the default
#                            1, that is no scaling. Note that only simple proportionality scaling 
#                            is possible. 
# $id->get_scale()           Returns the current scale factor.
# $id->set_limit($$)         Sets the plot limit of one of the fololowing tags:
#                            'xmin', xmax', 'ymin' and 'ymax' to the specified value.
#                            Arguments: tag, value.
# $id->get_limit($)          Returns the limit of the specified tag (see set_limit()).
# $id->set_xlabel($)         Explicitly sets the text string to be displayed in the xlabel.
# $id->set_ylabel($)         Explicitly sets the text string to be displayed in the ylabel.
# $id->set_ylabel_bode($)    Explicitly sets the text string to be displayed in the 
#                            ylabel in bode plots.
# $id->set_bode_mode($)      Sets wether bode plot should only plot imaginary value or both
#                            real and imaginary (default).
#                            If called with a true argument mode is set to only imaginary values.
# $id->set_admittance_mode($) Sets wether or not the plots should be of the impedance,
#                            admittance, complex modulus or compelx capacitance.
#                            Default is impedance. Valid values can be obtained from admittance_mode_values().
# $id->admittance_mode_values() Returns a list of valid values for set_admittance_mode($).
# $id->adv_opt([@])          Sets or gets a list of additional gnuplot commands (for instance
#                            user defined axis labels etc.). Arguments are a list of commands.
#                            If called without arguments, the current list of additional 
#                            commands is returned
#
# Additionally the following private member functions are also defined. Although
# Perl does not permit true private functions, do not use these functions from
# outside the class instances!
#
# $id->print_plot_common()   This function returns the function declarations for all 
#                            elements in the impedance model (note returns a list).
# $id->print_plot_main()     This function does the actual work of print_plot() and
#                            print_bode()
# $id->print_plot_main_range() This function does the actual work of print_plot_range(). 
# $id->print_plot_eps_common() Similar as above, bur for eps output.
# $id->dev_sort_fmax()       Returns a hash of device names and summit frequencies
#                            Only returns devices for which summit frequencies can be
#                            correctly calculated.
# $id->print_simulation()    This function does the actual work of print_imp_sim()
#                            and print_bode_sim().
# $id->save()                This function saves the content of the data field to
#                            a temporary file (in /temp) It returns the filename of
#                            the temporary file.
# $id->file()                Returns the name of the data file.
#
##############################################################
use strict;
use warnings;
use Math::Trig;
use Math::Complex;
use POSIX qw();
use Impedance::Device;
use SemaforeFile;
use Debug;
our $AUTOLOAD;
our @ISA = qw(Debug);
BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new
{
    my $class = shift;
    my $self = {
	'_devices' => {},
	'_data' => [],
	'_file' => undef,
	'_filename' => undef,
	'_fitset' => undef,
	'_show_arcs' => 0,
	'_dev_sort' => {},
	'_xmin' => undef,
	'_xmax' => undef,
	'_ymin' => undef,
	'_ymax' => undef,
	'_xlabel' => 'Z_real(Ohm)',
	'_ylabel' => '-Z_imag(Ohm)',
	'_ylabel_bode' => 'Z(Ohm)',
	'_scale_factor' => 1,
	'_bode_mode' => 0,
	'_plot_admittance' => 'Impedance',
	'_values_plot' => ['Impedance','Admittance','Complex modulus','Complex capacitance'],
	'_adv_opt' => [],
	'_unlinkfiles' => [],
    };
    bless $self,$class;
    if (@_)
    {
	my $in = shift;
	if (ref($in) eq 'Impedance::Model')
	{
	    if (@_)
	    {
		my $min = shift;
		my $max = shift;
		$self->data($in->subset($min,$max)) if ($in->data);
	    }
	    else
	    {
		$self->data($in->data) if ($in->data);
	    }
	    foreach my $e (sort keys %{$in->{'_devices'}})
	    {		
		my $type = $in->element($e)->type;
		my $id = $self->add_element($type);
		foreach my $k ($id->tags)
		{
		    $id->value($k,$in->element($e)->value($k));
		}
	    }
	}
	else
	{
	    $self->data($in);
	}
    }
    return $self;
}

sub admittance_mode_values()
{
    my $self = shift;
    return @{$self->{'_values_plot'}};
}

sub set_admittance_mode($)
{
    my $self = shift;
    $self->{'_plot_admittance'} = shift;
}

sub set_bode_mode($)
{
    my $self = shift;
    $self->{'_bode_mode'} = shift;
}

sub set_xlabel($)
{
    my $self = shift;
    $self->{'_xlabel'} = shift;
}

sub set_ylabel($)
{
    my $self = shift;
    $self->{'_ylabel'} = shift;
}

sub set_ylabel_bode($)
{
    my $self = shift;
    $self->{'_ylabel_bode'} = shift;
}

sub plot_label($)
{
    my $self = shift;
    my $scale = $self->get_scale();
    return 1 unless (defined($scale));
    my $re;
    my $im;
    if (@_ > 1)
    {
	$re = shift;
	$im = shift;
    }
    else
    {
	my $f = shift;
	my $z = $scale * $self->F($f);
	if ($self->{'_plot_admittance'})
	{
	    $re = Re(1/$z);
	    $im = Im(1/$z);	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    $re = Re($z*cplx(0,1)*2*pi*$f);
	    $im = Im($z*cplx(0,1)*2*pi*$f);
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    $re = Re(1/($z*cplx(0,1)*2*pi*$f));
	    $im = Im(1/($z*cplx(0,1)*2*pi*$f));
	}
	else
	{	
	    $re = Re($z);
	    $im = Im($z);
	}
    }
    my $xmin = $self->{'_xmin'};
    my $xmax = $self->{'_xmax'};
    my $ymin = $self->{'_ymin'};
    my $ymax = $self->{'_ymax'};
    return 0 if (defined($xmin) && $re * $scale <= $xmin);
    return 0 if (defined($xmax) && $re * $scale >= $xmax);
    return 0 if (defined($ymin) && -$im * $scale <= $ymin);
    return 0 if (defined($ymax) && -$im * $scale >= $ymax);
    return 1;
}

sub set_limit($$)
{
    my $self = shift;
    my $key = shift;
    foreach my $tag ('xmin','xmax','ymin','ymax')
    {
	next unless ($key && $key eq $tag);
	$self->{'_'.$tag} = shift;
	return;
    }
}

sub get_limit($)
{
    my $self = shift;
    my $key = shift;
    foreach my $tag ('xmin','xmax','ymin','ymax')
    {
	next unless ($key && $key eq $tag);
	return $self->{'_'.$tag};
    }
}

sub scale_factor
{
    my $self = shift;
    $self->{'_scale_factor'} = 1;
    if (@_)
    {
	my $val = shift;
	$val = 1 unless ($val && $val > 0);
	$self->{'_scale_factor'} = $val;
    }
    return $self->{'_scale_factor'};
}

sub get_scale()
{
    my $self = shift;
    return $self->{'_scale_factor'};
}

sub set_show_arcs($)
{
    my $self = shift;
    $self->{'_show_arcs'} = shift;
}

sub minr()
{
    my $self = shift;
    my $min = undef;
    foreach my $s (@{$self->{'_data'}})
    {
	$min = $$s[1] if (defined($$s[1]) && (!defined($min) || $$s[1] < $min));
    }
    return $min;
}

sub maxr()
{
    my $self = shift;
    my $max = undef;
    foreach my $s (@{$self->{'_data'}})
    {
	$max = $$s[1] if (defined($$s[1]) && (!defined($max) || $$s[1] > $max));
    }
    return $max;
}

sub mini()
{
    my $self = shift;
    my $min = undef;
    foreach my $s (@{$self->{'_data'}})
    {
	$min = $$s[2] if (defined($$s[2]) && (!defined($min) || $$s[2] < $min));
    }
    return $min;
}

sub maxi()
{
    my $self = shift;
    my $max = undef;
    foreach my $s (@{$self->{'_data'}})
    {
	$max = $$s[2] if (defined($$s[2]) && (!defined($max) || $$s[2] > $max));
    }
    return $max;
}

sub minf()
{
    my $self = shift;
    my $min = undef;
    foreach my $s (@{$self->{'_data'}})
    {
	$min = $$s[0] if (defined($$s[0]) && (!defined($min) || $$s[0] < $min));
    }
    return $min;
}

sub maxf()
{
    my $self = shift;
    my $max = undef;
    foreach my $s (@{$self->{'_data'}})
    {
	$max = $$s[0] if (defined($$s[0]) && (!defined($max) || $$s[0] > $max));
    }
    return $max;
}


sub fit()
{
    my $self = shift;
    my @list = ();
    foreach (sort {$a<=>$b} keys %{$self->{'_devices'}})
    {
        push @list,$self->{'_devices'}{$_}->fit();
    }
    return @list;
}

sub parse($)
{
    my $self = shift;
    my $m = shift;
    $m = '' unless ($m);
    my $x = 0;
    my @arr = ($self);
    foreach my $l (split(/\n/,$m))
    {
	next unless (defined($l));
        if ($l =~ /[\(\[]/)
        {
            my $type = undef;
	    ###########################################################
	    ### Handle container elements with intrinsic properties ###
	    ### The format is:                                      ###
	    ### Name: Par1=val1, Par2=val2 (                        ###
	    ### [other devices as normal container]                 ###
	    ### )                                                   ###
	    ###########################################################
	    if ($l =~ /^[\s\t]*(\w+)\:\s*([\w\,\:\+\-\s\=\.]+)/)
	    {
		$type = $1;
		my $args = $2;
		my $id = $arr[-1]->add_element($type,++$x);
		push @arr,$id;
		foreach my $e (split(/\,\s+/,$args))
		{
		    my ($tag,$val) = split(/\=/,$e);
		    $id->value($tag,$val);
		}
		next;		
	    }
	    ### Handle pure containers (serial and parallel) ###
            $type = 'Par' if ($l =~ /^[\s\t]*\(/);
            $type = 'Ser' if ($l =~ /^[\s\t]*\[/);
            my $id = $arr[-1]->add_element($type,++$x);
            push @arr,$id;
            next;
        }
        if ($l =~ /^[\s\t]*[\)\]]/)
        {
            pop @arr;
            $x++;
            next;
        }
	next unless ($l =~ /\w/);
        my ($t,$args) = split(/\:\s+/,$l);
        if ($t =~ /\w+/)
        {
            my $id = $arr[-1]->add_element($t,++$x);
            foreach my $e (split(/\,\s+/,$args))
            {
                my ($tag,$val) = split(/\=/,$e);
                $id->value($tag,$val);
            }
        }
    }
    return $x;
}

sub print_model()
{
    my $self = shift;
    my @l = ();
    foreach my $e ($self->elements)
    {
	push @l,$self->element($e)->print_line;
    }
    return join("\n",@l);
}

sub device_types()
{
    my $self = shift;
    return Impedance::Device::list_device_types();
}

sub elements()
{
    my $self = shift;
    my @l = ();
    foreach (sort {$a<=>$b} keys %{$self->{'_devices'}})
    {
	push @l,$_;
    }
    return @l;
}

sub elements_all()
{
    my $self = shift;
    my @devlist = ();
    foreach my $x (keys %{$self->{'_devices'}})
    {
	my $dev = $self->{'_devices'}{$x};
	push @devlist,$dev;
	if (ref($dev) =~ /Par|Ser|[a-zA-Z]+\_$/)
	{
	    foreach my $d ($dev->elements_all)
	    {
		push @devlist,$d;
	    }
	}
    }
    return @devlist;
}

sub model_text()
{
    my $self = shift;
    return 'Z = 0' unless (%{$self->{'_devices'}});
    my $res = 'Z = ';
    my @l = ();
    foreach (sort {$a<=>$b} keys %{$self->{'_devices'}})
    {
	push @l,$self->{'_devices'}{$_}->name();
    }
    $res .= join(' ',@l);
    return $res;
}

sub element($)
{
    my $self = shift;
    my $id = shift;
    return undef unless ($id);
    foreach my $d ($self->elements_all)
    {
	return $d if ($id == $d->nr);
    }
    return undef;
}

sub delete_element($)
{
    my $self = shift;
    my $id = shift;
    return undef unless ($id && exists($self->{'_devices'}{$id}));
    delete($self->{'_devices'}{$id});
    $self->{'_dev_sort'} = {};
    return undef;
}

sub add_element
{
    my $self = shift;
    my $id = undef;
    my $type = uc(shift);
    my $nr = 0;
    my @l = ();
    ### Find largest current element number ###
    foreach (sort keys %{$self->{'_devices'}})
    {
	$nr = $_ if ($_ > $nr);
    }
    $nr++;
    if (@_)
    {
	$nr = shift;
    }
    die "Element with number $nr already exists!" if (exists($self->{'_devices'}{$nr}));
    $id = Impedance::Device::new_device($type,$nr);
    $self->{'_devices'}{$nr} = $id;
    $self->{'_dev_sort'} = {};
    return $id;
}

sub print_plot_common()
{
    my $self = shift;
    my @l = ('z(x) = x','sf(x) = '.$self->{'_scale_factor'}.' * z(x)');
    my @list = ();
    foreach (sort {$a<=>$b} keys %{$self->{'_devices'}})
    {
        push @list,$self->{'_devices'}{$_}->function_name();
        push @l,$self->{'_devices'}{$_}->functions();
    }
    if (@list)
    {
        push @l,'f(x) = '.join(' + ',@list);
    }
    else
    {
        push @l,'f(x) = {0,0}';
    }
    push @l,'h(x,y) = ( (y==1) ? real(f(x)): ( (y==2) ? imag(f(x)):  ( 1/0 )))';
    return @l;
}

sub print_fit_weight
{
    my $self = shift;
    my $filename = $self->file;
    if (@_)
    {
	$filename = shift;
    }
    unless ($filename)
    {
	$filename = $self->save() if (@{$self->{'_data'}});
    }
    my @fitlist = @_;
    my @data = ();
    open F,'<',$filename or die "Could not open file '$filename' $!";
    while (<F>)
    {
	s/^[\s\t]*//;
	next unless (/\d+/);
	my ($frq,$re,$im) = split(/[\s\t]+/,$_);
	my $er = ($re * $re + $im * $im)**0.5; ## Weight factor ##
	push @data,join("\t",1,$frq,$re,$er);
	push @data,join("\t",2,$frq,$im,$er);
    }
    close F;
    my $fitset = SemaforeFile->new('/tmp/fitset_'.$$.'.dat');
    $self->{'_fitset'} = $fitset->filename;
    push @{$self->{'_unlinkfiles'}},$fitset->filename;
    $fitset->writeline(@data);
    die "No fitting parameters supplied!" unless (@fitlist);
    my @l = $self->print_plot_common();
    push @l,"fit h(x,y) '".$fitset->filename."' u 2:1:3:4 via ".join(',',@fitlist);
    foreach (sort keys %{$self->{'_devices'}})
    {
        push @l,$self->{'_devices'}{$_}->save();
    }
    return join("\n",@l);
}

sub print_fit
{
    my $self = shift;
    my $filename = $self->file;
    if (@_)
    {
	$filename = shift;
    }
    unless ($filename)
    {
	$filename = $self->save() if (@{$self->{'_data'}});
    }
    my @fitlist = @_;
    my @data = ();
    open F,'<',$filename or die "Could not open file '$filename' $!";
    while (<F>)
    {
	s/^[\s\t]*//;
	next unless (/\d+/);
	my ($frq,$re,$im) = split(/[\s\t]+/,$_);
	my $er = (($re * $re + $im * $im)**0.5)/20;  # 5% error for weight #
	push @data,join("\t",1,$frq,$re,$er);
	push @data,join("\t",2,$frq,$im,$er);
    }
    close F;
    my $fitset = SemaforeFile->new('/tmp/fitset_'.$$.'.dat');
    push @{$self->{'_unlinkfiles'}},$fitset->filename;
    $self->{'_fitset'} = $fitset->filename;
    $fitset->writeline(@data);
    die "No fitting parameters supplied!" unless (@fitlist);
    my @l = $self->print_plot_common();
    push @l,"fit h(x,y) '".$fitset->filename."' u 2:1:3:(1) via ".join(',',@fitlist);
    foreach (sort keys %{$self->{'_devices'}})
    {
        push @l,$self->{'_devices'}{$_}->save();
    }
    return join("\n",@l);
}


sub print_plot
{
    my $self = shift;
    return $self->print_plot_main(@_);
}

sub print_plot_range($$)
{
    my $self = shift;
    return $self->print_plot_main_range(@_);
}

sub print_bode
{
    my $self = shift;
    if (@_)
    {
	return $self->print_plot_main(@_,1);
    }
    else
    {
	return $self->print_plot_main(undef,1);
    }
}

sub dev_sort_fmax()
{
    my $self = shift;
    return %{$self->{'_dev_sort'}} if (%{$self->{'_dev_sort'}});
    my %hash = ();
    foreach my $dev (keys %{$self->{'_devices'}})
    {
	my $f = $self->{'_devices'}{$dev}->f_imax();
	next unless (defined($f)); ## Skip elements for which summit frequency can not be defined.
	$f = 1e200 unless ($f); ## Handle series resistors (Ohmic)
	$hash{$dev} = $f;
    }
    $self->{'_dev_sort'} = \%hash;
    return %hash;
}

sub adv_opt(@)
{
    my $self = shift;
    if (@_)
    {
	my @opt = @_;
	$self->{'_adv_opt'} = \@opt;
    }
    return @{$self->{'_adv_opt'}};
}

sub print_plot_main
{
    my $self = shift;
    my $filename = $self->file;
    my $title = '';
    if (@_)
    {
	$filename = shift;
	$title = $filename;
	if ($filename && $filename =~ /\/([\w\.\-\s]+)/)
	{
	    $title = $1;
	}
    }
    unless ($filename)
    {
	$title = 'Data';
	$filename = $self->save() if (@{$self->{'_data'}});
    }
    my $mode = 0;
    $mode = 1 if (@_);
    my $minf = 1e100;
    my $maxf = 1e-100;
    die "No filename specified!" unless ($filename);
    open FN,'<',$filename or die "Could not open file $filename $!";
    my @data = ();
    my $mini = 1e100;
    my $minb = 1e100;
    my $minm = 1e100;
    my $minc = 1e100;
    while (<FN>)
    {
	s/^[\s\t]*//;
	next unless (/([\d\.\eE\-\+]+)/);
	my ($f,$r,$i) = split(/[\s\t]+/); 
	next unless ($r > 0 || $i > 0);
	next unless ($f > 0);
	$minf = $f if $f < $minf;
	$maxf = $f if $f > $maxf;
	my $b = -$i/($i*$i+$r*$r);
	my $m = Im(cplx($r,$i)*cplx(0,1)*2*pi*$f);
	my $c = Im(1/(cplx($r,$i)*cplx(0,1)*2*pi*$f));
	$mini = abs($i) if (abs($i) < $mini && $i != 0);
	$minb = abs($b) if (abs($b) < $minb && $b != 0);
	$minm = abs($m) if (abs($m) < $minm && $m != 0);
	$minc = abs($c) if (abs($c) < $minc && $c != 0);
	push @data,$_;
    }
    close FN;
    my $dist = (log($maxf)/log(10) - log($minf)/log(10))/6;
    my $f = log($maxf)/log(10);
    my $labelnr = 0;
    my @l = $self->print_plot_common();
    push @l,$self->adv_opt();
    while ($f > log($minf)/log(10))
    {
	$f -= $dist;
	for (my $x = 0; $x < @data; $x++)
	{	    
	    my ($frq,$re,$im) = split(/[\s\t]+/,$data[$x]);
	    next if (log($frq)/log(10) > $f);
	    if ($self->{'_plot_admittance'} eq 'Admittance')
	    {
		my $z = cplx($re,$im);
		my $a = 1/$z;
		push @l,'set label '.(++$labelnr).' "-- '.sprintf("%.3g",$frq).'" at sf('.Re($a).'), sf('.Im($a).'), 0 left rotate' unless ($mode || !$self->plot_label($re,$im));		
	    }
	    elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	    {
		my $z = cplx($re,$im);
		my $a = $z*cplx(0,1)*2*pi*$frq;
		push @l,'set label '.(++$labelnr).' "-- '.POSIX::floor($x+0.5).'" at sf('.Re($a).'), sf('.(Im($a)).'), 0 left rotate' unless ($mode || !$self->plot_label($re,$im));
	    }
	    elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	    {
		my $z = cplx($re,$im);
		my $a = 1/($z*cplx(0,1)*2*pi*$frq);
		push @l,'set label '.(++$labelnr).' "-- '.POSIX::floor($x+0.5).'" at sf('.Re($a).'), sf('.(-Im($a)).'), 0 left rotate' unless ($mode || !$self->plot_label($re,$im));	    
	    }
	    else
	    {
		push @l,'set label '.(++$labelnr).' "-- '.sprintf("%.3g",$frq).'" at sf('.$re.'), sf('.-$im.'), 0 left rotate' unless ($mode || !$self->plot_label($re,$im));		
	    }
	    last;
	}
    }
    my $tit2 = $self->model_text;
    push @l,'# End Model #';
    push @l,'set nokey';
    if ($mode)
    {
	my $ylabel = $self->{'_ylabel_bode'};
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    $ylabel =~ s/\-Z/Y/; 
	    $ylabel =~ s/Ohm/S/; 
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    $ylabel =~ s/-Z/M/; 
	    $ylabel =~ s/Ohm/Ohm\/s/; 	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    $ylabel =~ s/-Z/C\*/; 
	    $ylabel =~ s/Ohm/F/; 	    	    
	}
        push @l,"set ylabel '".$ylabel."'";
        push @l,"set xlabel 'f(Hz)'";
        push @l,"set logscale x";
        push @l,"set logscale y";
        push @l,"p(x) = x>0?x:1/0";
        push @l,"n(x) = x<0?-x:1/0";
        push @l,'set format x "10^{%L}"';
        push @l,'set format y "10^{%L}"';
	if ($title =~ /Z = 0/)
	{
	    push @l,'set yrange [1:10]';
	    push @l,"set xrange [$minf:$maxf]";
	}
	else
	{
	    my $scale = $self->get_scale();
	    if ($self->{'_plot_admittance'} eq 'Admittance')
	    {
		my $minir = 10 ** POSIX::floor(log($minb)/log(10) - 1) * $scale;
		push @l,"set yrange [$minir:]";		
	    }
	    elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	    {
		my $minir = 10 ** POSIX::floor(log($minm)/log(10) - 1) * $scale;
		push @l,"set yrange [$minir:]";				
	    }
	    elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	    {
		my $minir = 10 ** POSIX::floor(log($minc)/log(10) - 1) * $scale;
		push @l,"set yrange [$minir:]";						
	    }	    
	    else
	    {
		my $minir = 10 ** POSIX::floor(log($mini)/log(10) - 1) * $scale;
		push @l,"set yrange [$minir:]";		
	    }
	}
        push @l,'#OUTPUT#';
	push @l,$self->adv_opt();
	my $plotstr;
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    push @l,'G(x,y) = x/(x*x+y*y)';
	    push @l,'B(x,y) = -y/(x*x+y*y)';
	    $plotstr = "plot '$filename' u 1:(sf(p(G(\$2,\$3)))) w p,'' u 1:(sf(n(G(\$2,\$3)))) w p,'' u 1:(sf(n(B(\$2,\$3)))) w p,'' u 1:(sf(p(B(\$2,\$3)))) w p, '' u 1:(sf(p(real(1/f(\$1))))) w l lt 2,'' u 1:(sf(n(real(1/f(\$1))))) w l lt 1,'' u 1:(sf(p(imag(1/f(\$1))))) w l lt 1,'' u 1:(sf(n(imag(1/f(\$1))))) w l lt 2";	
	    $plotstr = "plot '$filename' u 1:(sf(n(B(\$2,\$3)))) w p,'' u 1:(sf(p(B(\$2,\$3)))) w p,'' u 1:(sf(p(imag(1/f(\$1))))) w l lt 1,'' u 1:(sf(n(imag(1/f(\$1))))) w l lt 2" if ($self->{'_bode_mode'});	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * x';
	    push @l,'Z(x,y) = {1,0} * x + {0,1} * y';
	    $plotstr = "plot '$filename' u 1:(sf(p(real(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(n(real(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(n(imag(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(p(imag(Z(\$2,\$3)*M(\$1))))) w p, '' u 1:(sf(p(real(f(\$1)*M(\$1))))) w l lt 2,'' u 1:(sf(n(real(f(\$1)*M(\$1))))) w l lt 1,'' u 1:(sf(p(imag(f(\$1)*M(\$1))))) w l lt 1,'' u 1:(sf(n(imag(f(\$1)*M(\$1))))) w l lt 2";	
	    $plotstr = "plot '$filename' u 1:(sf(n(imag(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(p(imag(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(p(imag(f(\$1)*M(\$1))))) w l lt 1,'' u 1:(sf(n(imag(f(\$1)*M(\$1))))) w l lt 2" if ($self->{'_bode_mode'});	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * x';
	    push @l,'Z(x,y) = {1,0} * x + {0,1} * y';
	    $plotstr = "plot '$filename' u 1:(sf(p(real(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(n(real(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(n(imag(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(p(imag(1/(Z(\$2,\$3)*M(\$1)))))) w p, '' u 1:(sf(p(real(1/(f(\$1)*M(\$1)))))) w l lt 2,'' u 1:(sf(n(real(1/(f(\$1)*M(\$1)))))) w l lt 1,'' u 1:(sf(p(imag(1/(f(\$1)*M(\$1)))))) w l lt 1,'' u 1:(sf(n(imag(1/(f(\$1)*M(\$1)))))) w l lt 2";	
	    $plotstr = "plot '$filename' u 1:(sf(n(imag(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(p(imag(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(p(imag(1/(f(\$1)*M(\$1)))))) w l lt 1,'' u 1:(sf(n(imag(1/(f(\$1)*M(\$1)))))) w l lt 2" if ($self->{'_bode_mode'});	    	    
	}
	else
	{	
	    $plotstr = "plot '$filename' u 1:(sf(p(\$2))) w p,'' u 1:(sf(n(\$2))) w p,'' u 1:(sf(n(\$3))) w p,'' u 1:(sf(p(\$3))) w p, '' u 1:(sf(p(real(f(\$1))))) w l lt 2,'' u 1:(sf(n(real(f(\$1))))) w l lt 1,'' u 1:(sf(p(imag(f(\$1))))) w l lt 1,'' u 1:(sf(n(imag(f(\$1))))) w l lt 2";	
	    $plotstr = "plot '$filename' u 1:(sf(n(\$3))) w p,'' u 1:(sf(p(\$3))) w p,'' u 1:(sf(p(imag(f(\$1))))) w l lt 1,'' u 1:(sf(n(imag(f(\$1))))) w l lt 2" if ($self->{'_bode_mode'});
	}
	my $lt = 3;
	my %hash = $self->dev_sort_fmax();
	if ($self->{'_show_arcs'})
	{
	    foreach my $dev (sort { $hash{$b} <=> $hash{$a} } (keys %hash))
	    {
		my $fn = $self->{'_devices'}{$dev}->function_name();
		$fn =~ s/\(x\)//;
		if ($self->{'_plot_admittance'} eq 'Admittance')
		{
#                       Does it even make sense to plot subarcs for admittance???? ####
		    $plotstr .= ",'' u 1:(abs(imag(sf(1/$fn(\$1))))) w l lt ".$lt++;
		}
		elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
		{
		    $plotstr .= ",'' u 1:(abs(imag(sf($fn(\$1)*M(\$1))))) w l lt ".$lt++;
		}
		elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
		{
		    $plotstr .= ",'' u 1:(abs(imag(sf(1/($fn(\$1)*M(\$1)))))) w l lt ".$lt++;
		}
		else
		{
		    $plotstr .= ",'' u 1:(abs(imag(sf($fn(\$1))))) w l lt ".$lt++;
		}
	    }  
	}
	push @l,$plotstr;
    }
    else
    {
	push @l,'set size ratio -1';
	my $xlabel = $self->{'_xlabel'};
	my $ylabel = $self->{'_ylabel'};
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    $xlabel =~ s/Z/Y/;
	    $ylabel =~ s/\-Z/Y/; 
	    $xlabel =~ s/Ohm/S/;
	    $ylabel =~ s/Ohm/S/; 
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    $xlabel =~ s/Z/M/;
	    $ylabel =~ s/\-Z/M/; 
	    $xlabel =~ s/Ohm/Ohm\/s/;
	    $ylabel =~ s/Ohm/Ohm\/s/; 
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    $xlabel =~ s/Z/C\*/;
	    $ylabel =~ s/\-Z/C\*/; 
	    $xlabel =~ s/Ohm/F/;
	    $ylabel =~ s/Ohm/F/; 
	}
	push @l,"set ylabel '".$ylabel."'";
	push @l,"set xlabel '".$xlabel."'";
	push @l,'#OUTPUT#';
	push @l,$self->adv_opt();
	my $lt = 3;
	my $plotstr;
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    push @l,'G(x,y) = x/(x*x+y*y)';
	    push @l,'B(x,y) = -y/(x*x+y*y)';
	    $plotstr = "plot '$filename' u (sf(G(\$2,\$3))):(sf(B(\$2,\$3))) w p,'' u (real(sf(1/f(\$1)))):(imag(sf(1/f(\$1)))) w l lt 2";
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * x';
	    push @l,'Z(x,y) = {1,0} * x + {0,1} * y';
	    $plotstr = "plot '$filename' u (sf(real(Z(\$2,\$3)*M(\$1)))):(sf(imag(Z(\$2,\$3)*M(\$1)))) w p,'' u (real(sf(f(\$1)*M(\$1)))):(imag(sf(f(\$1)*M(\$1)))) w l lt 2";
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * x';
	    push @l,'Z(x,y) = {1,0} * x + {0,1} * y';
	    $plotstr = "plot '$filename' u (sf(real(1/(Z(\$2,\$3)*M(\$1))))):(sf(-imag(1/(Z(\$2,\$3)*M(\$1))))) w p,'' u (real(sf(1/(f(\$1)*M(\$1))))):(-imag(sf(1/(f(\$1)*M(\$1))))) w l lt 2";
	}
	else
	{
	    $plotstr = "plot '$filename' u (sf(\$2)):(sf(-\$3)) w p,'' u (real(sf(f(\$1)))):(-imag(sf(f(\$1)))) w l lt 2";
	    if ($self->{'_show_arcs'})
	    {
		my @list = ();
		my %hash = $self->dev_sort_fmax();
		foreach my $dev (sort { $hash{$b} <=> $hash{$a} } (keys %hash))
		{
		    my $fn = $self->{'_devices'}{$dev}->function_name();
		    $fn =~ s/\(x\)//;
		    my $offsetx = 0;
		    foreach my $d (@list)
		    {
			$offsetx +=  Re($self->{'_devices'}{$d}->F(1e-100));
		    }
		  OUTER:
		    foreach my $key (keys %{$self->{'_devices'}})
		    {
			foreach (@list)
			{
			    next OUTER if ($_ eq $key);
			}
			$offsetx += Re($self->{'_devices'}{$key}->F(1e100));
		    }
		    my $diff = Re($self->{'_devices'}{$dev}->F(1e-100) - $self->{'_devices'}{$dev}->F(1e100));
		    $offsetx += $diff if ($diff < 0);
		    $plotstr .= ",'' u (sf($offsetx + real($fn(\$1)))):(-imag(sf($fn(\$1)))) w l lt ".$lt++;
		    push @list,$dev;
		}  
	    }
	}
	push @l,$plotstr;
    }
    return join("\n",@l);
}

sub print_plot_main_range
{
    my $self = shift;
    my $minx = shift;
    my $maxx = shift;
    my $filename = $self->file;
    my $title = '';
    if (@_)
    {
	$filename = shift;
	$title = $filename;
	if ($filename && $filename =~ /\/([\w\.\-\s]+)/)
	{
	    $title = $1;
	}
    }
    unless ($filename)
    {
	$title = 'Data';
	$filename = $self->save() if (@{$self->{'_data'}});
    }
    my $mode = 0;
    $mode = 1 if (@_);
    my $minf = 1e100;
    my $maxf = 1e-100;
    die "No filename specified!" unless ($filename);
    open FN,'<',$filename or die "Could not open file $filename $!";
    my @data = ();
    my $mini = 1e100;
    my $minb = 1e100;
    my $minm = 1e100;
    my $minc = 1e100;
    while (<FN>)
    {
	s/^[\s\t]*//;
	next unless (/([\d\.\eE\-\+]+)/);
	my ($f,$r,$i) = split(/[\s\t]+/); 
	$minf = $f if $f < $minf;
	$maxf = $f if $f > $maxf;
	my $b = -$i/($i*$i+$r*$r);
	my $m = Im(cplx($r,$i)*cplx(0,1)*2*pi*$f);
	my $c = Im(1/(cplx($r,$i)*cplx(0,1)*2*pi*$f));
	$mini = abs($i) if (abs($i) < $mini && $i != 0);
	$minb = abs($b) if (abs($b) < $minb && $b != 0);
	$minm = abs($m) if (abs($m) < $minm && $m != 0);
	$minc = abs($c) if (abs($c) < $minc && $c != 0);
	push @data,$_;
    }
    close FN;
    my $dist = (log($maxf)/log(10) - log($minf)/log(10))/6;
    my $f = log($maxf)/log(10);
    my $labelnr = 0;
    my @l = $self->print_plot_common();
    push @l,$self->adv_opt();
    while ($f > log($minf)/log(10))
    {
	$f -= $dist;
	for (my $x = 0; $x < @data; $x++)
	{
	    
	    my ($frq,$re,$im) = split(/[\s\t]+/,$data[$x]);
	    next if (log($frq)/log(10) > $f);
	    next if ($re < $minx || $re > $maxx);
	    if ($self->{'_plot_admittance'} eq 'Admittance')
	    {
		my $z = cplx($re,$im);
		my $a = 1/$z;
		push @l,'set label '.(++$labelnr).' "-- '.sprintf("%.3g",$frq).'" at sf('.Re($a).'), sf('.Im($a).'), 0 left rotate' unless ($mode || !$self->plot_label($re,$im));		
	    }
	    elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	    {
		my $z = cplx($re,$im);
		my $a = $z*cplx(0,1)*2*pi*$frq;
		push @l,'set label '.(++$labelnr).' "-- '.POSIX::floor($x+0.5).'" at sf('.Re($a).'), sf('.(Im($a)).'), 0 left rotate' unless ($mode || !$self->plot_label($re,$im));
	    }
	    elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	    {
		my $z = cplx($re,$im);
		my $a = 1/($z*cplx(0,1)*2*pi*$frq);
		push @l,'set label '.(++$labelnr).' "-- '.POSIX::floor($x+0.5).'" at sf('.Re($a).'), sf('.(-Im($a)).'), 0 left rotate' unless ($mode || !$self->plot_label($re,$im));	    
	    }
	    else
	    {
		push @l,'set label '.(++$labelnr).' "-- '.sprintf("%.3g",$frq).'" at sf('.$re.'), sf('.(-$im).'), 0 left rotate' unless ($mode || !$self->plot_label($re,$im));
	    }
	    last;
	}
    }
    my $tit2 = $self->model_text;
    push @l,'# End Model #';
    push @l,'set nokey';
    if ($mode)
    {
	my $ylabel = $self->{'_ylabel_bode'};
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    $ylabel =~ s/\-Z/Y/; 
	    $ylabel =~ s/Ohm/S/; 
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    $ylabel =~ s/-Z/M/; 
	    $ylabel =~ s/Ohm/Ohm\/s/; 	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    $ylabel =~ s/-Z/C\*/; 
	    $ylabel =~ s/Ohm/F/; 	    	    
	}
        push @l,"set ylabel '".$ylabel."'";
        push @l,"set xlabel 'f(Hz)'";
        push @l,"set logscale x";
        push @l,"set logscale y";
        push @l,"p(x) = x>0?x:1/0";
        push @l,"n(x) = x<0?-x:1/0";
        push @l,'set format x "10^{%L}"';
        push @l,'set format y "10^{%L}"';
	if ($title =~ /Z = 0/)
	{
	    push @l,'set yrange [1:10]';
	    push @l,"set xrange [$minf:$maxf]";
	}
	else
	{
	    my $scale = $self->get_scale();
	    if ($self->{'_plot_admittance'} eq 'Admittance')
	    {
		my $minir = 10 ** POSIX::floor(log($minb)/log(10) - 1) * $scale;
		push @l,"set yrange [$minir:]";		
	    }
	    elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	    {
		my $minir = 10 ** POSIX::floor(log($minm)/log(10) - 1) * $scale;
		push @l,"set yrange [$minir:]";				
	    }
	    elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	    {
		my $minir = 10 ** POSIX::floor(log($minc)/log(10) - 1) * $scale;
		push @l,"set yrange [$minir:]";						
	    }	    
	    else
	    {
		my $minir = 10 ** POSIX::floor(log($mini)/log(10) - 1) * $scale;
		push @l,"set yrange [$minir:]";	    
	    }
	}
        push @l,'#OUTPUT#';
	push @l,$self->adv_opt();
	my $lt = 3;
	my $plotstr;
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    push @l,'G(x,y) = x/(x*x+y*y)';
	    push @l,'B(x,y) = -y/(x*x+y*y)';
	    $plotstr = "plot '$filename' u 1:(sf(p(G(\$2,\$3)))) w p,'' u 1:(sf(n(G(\$2,\$3)))) w p,'' u 1:(sf(n(B(\$2,\$3)))) w p,'' u 1:(sf(p(B(\$2,\$3)))) w p, '' u 1:(sf(p(real(1/f(\$1))))) w l lt 2,'' u 1:(sf(n(real(1/f(\$1))))) w l lt 1,'' u 1:(sf(p(imag(1/f(\$1))))) w l lt 1,'' u 1:(sf(n(imag(1/f(\$1))))) w l lt 2";	
	    $plotstr = "plot '$filename' u 1:(sf(n(B(\$2,\$3)))) w p,'' u 1:(sf(p(B(\$2,\$3)))) w p,'' u 1:(sf(p(imag(1/f(\$1))))) w l lt 1,'' u 1:(sf(n(imag(1/f(\$1))))) w l lt 2" if ($self->{'_bode_mode'});	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * x';
	    push @l,'Z(x,y) = {1,0} * x + {0,1} * y';
	    $plotstr = "plot '$filename' u 1:(sf(p(real(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(n(real(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(n(imag(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(p(imag(Z(\$2,\$3)*M(\$1))))) w p, '' u 1:(sf(p(real(f(\$1)*M(\$1))))) w l lt 2,'' u 1:(sf(n(real(f(\$1)*M(\$1))))) w l lt 1,'' u 1:(sf(p(imag(f(\$1)*M(\$1))))) w l lt 1,'' u 1:(sf(n(imag(f(\$1)*M(\$1))))) w l lt 2";	
	    $plotstr = "plot '$filename' u 1:(sf(n(imag(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(p(imag(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(p(imag(f(\$1)*M(\$1))))) w l lt 1,'' u 1:(sf(n(imag(f(\$1)*M(\$1))))) w l lt 2" if ($self->{'_bode_mode'});	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * x';
	    push @l,'Z(x,y) = {1,0} * x + {0,1} * y';
	    $plotstr = "plot '$filename' u 1:(sf(p(real(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(n(real(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(n(imag(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(p(imag(1/(Z(\$2,\$3)*M(\$1)))))) w p, '' u 1:(sf(p(real(1/(f(\$1)*M(\$1)))))) w l lt 2,'' u 1:(sf(n(real(1/(f(\$1)*M(\$1)))))) w l lt 1,'' u 1:(sf(p(imag(1/(f(\$1)*M(\$1)))))) w l lt 1,'' u 1:(sf(n(imag(1/(f(\$1)*M(\$1)))))) w l lt 2";	
	    $plotstr = "plot '$filename' u 1:(sf(n(imag(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(p(imag(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(p(imag(1/(f(\$1)*M(\$1)))))) w l lt 1,'' u 1:(sf(n(imag(1/(f(\$1)*M(\$1)))))) w l lt 2" if ($self->{'_bode_mode'});	    	    
	}
	else
	{	
	    $plotstr = "plot '$filename' u 1:(sf(p(\$2))) w p,'' u 1:(sf(n(\$2))) w p,'' u 1:(sf(n(\$3))) w p,'' u 1:(sf(p(\$3))) w p, '' u 1:(sf(p(real(f(\$1))))) w l lt 2,'' u 1:(sf(n(real(f(\$1))))) w l lt 1,'' u 1:(sf(p(imag(f(\$1))))) w l lt 2,'' u 1:(sf(n(imag(f(\$1))))) w l lt 1";
	    $plotstr = "plot '$filename'  u 1:(sf(n(\$3))) w p,'' u 1:(sf(p(\$3))) w p,'' u 1:(sf(p(imag(f(\$1))))) w l lt 2,'' u 1:(sf(n(imag(f(\$1))))) w l lt 1" if ($self->{'_bode_mode'});
	}
	if ($self->{'_show_arcs'})
	{
	    my %hash = $self->dev_sort_fmax();
	    foreach my $dev (sort { $hash{$b} <=> $hash{$a} } (keys %hash))
	    {
		my $fn = $self->{'_devices'}{$dev}->function_name();
		$fn =~ s/\(x\)//;
		if ($self->{'_plot_admittance'} eq 'Admittance')
		{
#                   Does it even make sense to plot subarcs for admittance???? ####
		    $plotstr .= ",'' u 1:(abs(imag(sf(1/$fn(\$1))))) w l lt ".$lt++;
		}
		elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
		{
		    $plotstr .= ",'' u 1:(abs(imag(sf($fn(\$1)*M(\$1))))) w l lt ".$lt++;
		}
		elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
		{
		    $plotstr .= ",'' u 1:(abs(imag(sf(1/($fn(\$1)*M(\$1)))))) w l lt ".$lt++;
		}
		else
		{
		    $plotstr .= ",'' u 1:(sf(abs(imag($fn(\$1))))) w l lt ".$lt++;
		}
	    }
	}
	push @l,$plotstr;
    }
    else
    {
	push @l,'set size ratio -1';
	my $xlabel = $self->{'_xlabel'};
	my $ylabel = $self->{'_ylabel'};
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    $xlabel =~ s/Z/Y/;
	    $ylabel =~ s/\-Z/Y/; 
	    $xlabel =~ s/Ohm/S/;
	    $ylabel =~ s/Ohm/S/; 
	}	
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    $xlabel =~ s/Z/M/;
	    $ylabel =~ s/\-Z/M/; 
	    $xlabel =~ s/Ohm/Ohm\/s/;
	    $ylabel =~ s/Ohm/Ohm\/s/; 
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    $xlabel =~ s/Z/C\*/;
	    $ylabel =~ s/\-Z/C\*/; 
	    $xlabel =~ s/Ohm/F/;
	    $ylabel =~ s/Ohm/F/; 
	}
	push @l,"set ylabel '".$ylabel."'";
	push @l,"set xlabel '".$xlabel."'";
	push @l,'#OUTPUT#';
	push @l,$self->adv_opt();
	my $plotstr;
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    push @l,'G(x,y) = x/(x*x+y*y)';
	    push @l,'B(x,y) = -y/(x*x+y*y)';
	    $plotstr = "plot '$filename' u (sf(G(\$2,\$3))):(sf(B(\$2,\$3))) w p,'' u (real(sf(1/f(\$1)))):(imag(sf(1/f(\$1)))) w l lt 2";
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * x';
	    push @l,'Z(x,y) = {1,0} * x + {0,1} * y';
	    $plotstr = "plot '$filename' u (sf(real(Z(\$2,\$3)*M(\$1)))):(sf(imag(Z(\$2,\$3)*M(\$1)))) w p,'' u (real(sf(f(\$1)*M(\$1)))):(imag(sf(f(\$1)*M(\$1)))) w l lt 2";
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * x';
	    push @l,'Z(x,y) = {1,0} * x + {0,1} * y';
	    $plotstr = "plot '$filename' u (sf(real(1/(Z(\$2,\$3)*M(\$1))))):(sf(-imag(1/(Z(\$2,\$3)*M(\$1))))) w p,'' u (real(sf(1/(f(\$1)*M(\$1))))):(-imag(sf(1/(f(\$1)*M(\$1))))) w l lt 2";
	}
	else
	{
	    $plotstr = "plot '$filename' u (sf(\$2)):(sf(-\$3)) w p,'' u (real(sf(f(\$1)))):(-imag(sf(f(\$1)))) w l lt 1";
	    my $lt = 3;
	    if ($self->{'_show_arcs'})
	    {
		my @list = ();
		my %hash = $self->dev_sort_fmax();
		foreach my $dev (sort { $hash{$b} <=> $hash{$a} } (keys %hash))
		{
		    my $fn = $self->{'_devices'}{$dev}->function_name();
		    $fn =~ s/\(x\)//;
		    my $offsetx = 0;
		    foreach my $d (@list)
		    {
			$offsetx += Re($self->{'_devices'}{$d}->F(1e-7));
		    }
		  OUTER:
		    foreach my $key (keys %{$self->{'_devices'}})
		    {
			foreach (@list)
			{
			    next OUTER if ($_ eq $key);
			}
			$offsetx += Re($self->{'_devices'}{$key}->F(1e100));
		    }
		    my $diff = Re($self->{'_devices'}{$dev}->F(1e-100) - $self->{'_devices'}{$dev}->F(1e100));
		    $offsetx += $diff if ($diff < 0);
		    $plotstr .= ",'' u (sf($offsetx + real($fn(\$1)))):(sf(-imag($fn(\$1)))) w l lt ".$lt++;
		    push @list,$dev;
		}  
	    }
	}
	push @l,$plotstr;
    }
    return join("\n",@l);
}

sub print_plot_eps
{
    my $self = shift;
    return $self->print_plot_eps_common(@_);
}

sub print_bode_eps
{
    my $self = shift;
    if (@_)
    {
	return $self->print_plot_eps_common(@_,1);
    }
    else
    {
	return $self->print_plot_eps_common(undef,1);
    }
}

sub print_plot_eps_common
{
    my $self = shift;
    my $filename = $self->file;
    my $title = '';
    if (@_)
    {
	$filename = shift;
	$title = $filename;
	if ($filename && $filename =~ /\/([\w\.\-\s]+)/)
	{
	    $title = $1;
	}
    }
    my $mode = 0;
    $mode = 1 if (@_);
    unless ($filename)
    {
	$title = 'Data';
	$filename = $self->save() if (@{$self->{'_data'}});
    }
    my $minf = 1e100;
    my $maxf = 1e-100;
    die "No filename specified!" unless ($filename);
    open FN,'<',$filename or die "Could not open file $filename $!";
    my @data = ();
    my $mini = 1e100;
    my $minb = 1e100;
    my $minm = 1e100;
    my $minc = 1e100;
    while (<FN>)
    {
	s/^[\s\t]*//;
	next unless (/([\d\.\eE\-\+]+)/);
	my ($f,$r,$i) = split(/[\s\t]+/); 
	$minf = $f if $f < $minf;
	$maxf = $f if $f > $maxf;
	my $b = -$i/($i*$i+$r*$r);
	my $m = Im(cplx($r,$i)*cplx(0,1)*2*pi*$f);
	my $c = Im(1/(cplx($r,$i)*cplx(0,1)*2*pi*$f));
	$mini = abs($i) if (abs($i) < $mini && $i != 0);
	$minb = abs($b) if (abs($b) < $minb && $b != 0);
	$minm = abs($m) if (abs($m) < $minm && $m != 0);
	$minc = abs($c) if (abs($c) < $minc && $c != 0);
	push @data,$_;
    }
    close FN;
    my $dist = (log($maxf)/log(10) - log($minf)/log(10))/6;
    my $f = log($maxf)/log(10);
    my $labelnr = 0;
    my @l = $self->print_plot_common();
    push @l,$self->adv_opt();
    while ($f > log($minf)/log(10))
    {
	$f -= $dist;
	for (my $x = 0; $x < @data; $x++)
	{
	    
	    my ($frq,$re,$im) = split(/[\s\t]+/,$data[$x]);
	    next if (log($frq)/log(10) > $f);
	    if ($self->{'_plot_admittance'} eq 'Admittance')
	    {
		my $z = cplx($re,$im);
		my $a = 1/$z;
		push @l,'set label '.(++$labelnr).' "-- '.sprintf("%.3g",$frq).'" at sf('.Re($a).'), sf('.Im($a).'), 0 left rotate' unless ($mode || !$self->plot_label($re,$im));		
	    }
	    elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	    {
		my $z = cplx($re,$im);
		my $a = $z*cplx(0,1)*2*pi*$frq;
		push @l,'set label '.(++$labelnr).' "-- '.POSIX::floor($x+0.5).'" at sf('.Re($a).'), sf('.(Im($a)).'), 0 left rotate' unless ($mode || !$self->plot_label($re,$im));
	    }
	    elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	    {
		my $z = cplx($re,$im);
		my $a = 1/($z*cplx(0,1)*2*pi*$frq);
		push @l,'set label '.(++$labelnr).' "-- '.POSIX::floor($x+0.5).'" at sf('.Re($a).'), sf('.(-Im($a)).'), 0 left rotate' unless ($mode || !$self->plot_label($re,$im));	    
	    }
	    else
	    {
		push @l,'set label '.(++$labelnr).' "-- '.sprintf("%.3g",$frq).'" at sf('.$re.'), sf('.(-$im).'), 0 left rotate' unless ($mode || !$self->plot_label($re,$im));		
	    }
	    last;
	}
    }
    my $tit2 = $self->model_text;
    push @l,'# End Model #';
    push @l,'set nokey';
    if ($mode)
    {
	my $ylabel = $self->{'_ylabel_bode'};
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    $ylabel =~ s/\-Z/Y/; 
	    $ylabel =~ s/Ohm/S/; 
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    $ylabel =~ s/-Z/M/; 
	    $ylabel =~ s/Ohm/Ohm\/s/; 	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    $ylabel =~ s/-Z/C\*/; 
	    $ylabel =~ s/Ohm/F/; 	    	    
	}
        push @l,"set ylabel '".$ylabel."'";
        push @l,"set xlabel 'f(Hz)'";
        push @l,"set logscale x";
        push @l,"set logscale y";
        push @l,'set format x "10^{%L}"';
        push @l,'set format y "10^{%L}"';
        push @l,"p(x) = x>0?x:1/0";
        push @l,"n(x) = x<0?-x:1/0";
	if ($title =~ /Z = 0/)
	{
	    push @l,'set yrange [1:10]';
	    push @l,"set xrange [$minf:$maxf]";
	}
	else
	{
	    my $scale = $self->get_scale();
	    if ($self->{'_plot_admittance'} eq 'Admittance')
	    {
		my $minir = 10 ** POSIX::floor(log($minb)/log(10) - 1) * $scale;
		push @l,"set yrange [$minir:]";		
	    }
	    elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	    {
		my $minir = 10 ** POSIX::floor(log($minm)/log(10) - 1) * $scale;
		push @l,"set yrange [$minir:]";				
	    }
	    elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	    {
		my $minir = 10 ** POSIX::floor(log($minc)/log(10) - 1) * $scale;
		push @l,"set yrange [$minir:]";						
	    }	    
	    else
	    {
		my $minir = 10 ** POSIX::floor(log($mini)/log(10) - 1) * $scale;
		push @l,"set yrange [$minir:]";		
	    }
	}
        push @l,'#OUTPUT#';
	push @l,$self->adv_opt();
	my $plotstr;
	my $lt = 3;
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    push @l,'G(x,y) = x/(x*x+y*y)';
	    push @l,'B(x,y) = -y/(x*x+y*y)';
	    $plotstr = "plot '$filename' u 1:(sf(p(G(\$2,\$3)))) w p,'' u 1:(sf(n(G(\$2,\$3)))) w p,'' u 1:(sf(n(B(\$2,\$3)))) w p,'' u 1:(sf(p(B(\$2,\$3)))) w p, '' u 1:(sf(p(real(1/f(\$1))))) w l lt 2,'' u 1:(sf(n(real(1/f(\$1))))) w l lt 1,'' u 1:(sf(p(imag(1/f(\$1))))) w l lt 1,'' u 1:(sf(n(imag(1/f(\$1))))) w l lt 2";	
	    $plotstr = "plot '$filename' u 1:(sf(n(B(\$2,\$3)))) w p,'' u 1:(sf(p(B(\$2,\$3)))) w p,'' u 1:(sf(p(imag(1/f(\$1))))) w l lt 1,'' u 1:(sf(n(imag(1/f(\$1))))) w l lt 2" if ($self->{'_bode_mode'});	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * x';
	    push @l,'Z(x,y) = {1,0} * x + {0,1} * y';
	    $plotstr = "plot '$filename' u 1:(sf(p(real(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(n(real(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(n(imag(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(p(imag(Z(\$2,\$3)*M(\$1))))) w p, '' u 1:(sf(p(real(f(\$1)*M(\$1))))) w l lt 2,'' u 1:(sf(n(real(f(\$1)*M(\$1))))) w l lt 1,'' u 1:(sf(p(imag(f(\$1)*M(\$1))))) w l lt 1,'' u 1:(sf(n(imag(f(\$1)*M(\$1))))) w l lt 2";	
	    $plotstr = "plot '$filename' u 1:(sf(n(imag(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(p(imag(Z(\$2,\$3)*M(\$1))))) w p,'' u 1:(sf(p(imag(f(\$1)*M(\$1))))) w l lt 1,'' u 1:(sf(n(imag(f(\$1)*M(\$1))))) w l lt 2" if ($self->{'_bode_mode'});	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * x';
	    push @l,'Z(x,y) = {1,0} * x + {0,1} * y';
	    $plotstr = "plot '$filename' u 1:(sf(p(real(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(n(real(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(n(imag(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(p(imag(1/(Z(\$2,\$3)*M(\$1)))))) w p, '' u 1:(sf(p(real(1/(f(\$1)*M(\$1)))))) w l lt 2,'' u 1:(sf(n(real(1/(f(\$1)*M(\$1)))))) w l lt 1,'' u 1:(sf(p(imag(1/(f(\$1)*M(\$1)))))) w l lt 1,'' u 1:(sf(n(imag(1/(f(\$1)*M(\$1)))))) w l lt 2";	
	    $plotstr = "plot '$filename' u 1:(sf(n(imag(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(p(imag(1/(Z(\$2,\$3)*M(\$1)))))) w p,'' u 1:(sf(p(imag(1/(f(\$1)*M(\$1)))))) w l lt 1,'' u 1:(sf(n(imag(1/(f(\$1)*M(\$1)))))) w l lt 2" if ($self->{'_bode_mode'});	    	    
	}
	else
	{	
	    $plotstr = "plot '$filename' u 1:(sf(p(\$2))) w p,'' u 1:(sf(n(\$2))) w p,'' u 1:(sf(n(\$3))) w p,'' u 1:(sf(p(\$3))) w p, '' u 1:(sf(p(real(f(\$1))))) w l lt 2,'' u 1:(sf(n(real(f(\$1))))) w l lt 1,'' u 1:(sf(p(imag(f(\$1))))) w l lt 1,'' u 1:(sf(n(imag(f(\$1))))) w l lt 2";
	    $plotstr = "plot '$filename'  u 1:(sf(n(\$3))) w p,'' u 1:(sf(p(\$3))) w p,'' u 1:(sf(p(imag(f(\$1))))) w l lt 2,'' u 1:(sf(n(imag(f(\$1))))) w l lt 1" if ($self->{'_bode_mode'});
	}
	if ($self->{'_show_arcs'})
	{
	    my %hash = $self->dev_sort_fmax();
	    foreach my $dev (sort { $hash{$b} <=> $hash{$a} } (keys %hash))
	    {
		my $fn = $self->{'_devices'}{$dev}->function_name();
		$fn =~ s/\(x\)//;
		if ($self->{'_plot_admittance'} eq 'Admittance')
		{
#                       Does it even make sense to plot subarcs for admittance???? ####
		    $plotstr .= ",'' u 1:(abs(imag(sf(1/$fn(\$1))))) w l lt ".$lt++;
		}
		elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
		{
		    $plotstr .= ",'' u 1:(abs(imag(sf($fn(\$1)*M(\$1))))) w l lt ".$lt++;
		}
		elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
		{
		    $plotstr .= ",'' u 1:(abs(imag(sf(1/($fn(\$1)*M(\$1)))))) w l lt ".$lt++;
		}
		else
		{
		    $plotstr .= ",'' u 1:(abs(imag(sf($fn(\$1))))) w l lt ".$lt++;
		}
	    }
	}
	push @l,$plotstr;
    }
    else
    {
	push @l,'set size ratio -1';
	my $xlabel = $self->{'_xlabel'};
	my $ylabel = $self->{'_ylabel'};
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    $xlabel =~ s/Z/Y/;
	    $ylabel =~ s/\-Z/Y/; 
	    $xlabel =~ s/Ohm/S/;
	    $ylabel =~ s/Ohm/S/; 
	}	
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    $xlabel =~ s/Z/M/;
	    $ylabel =~ s/\-Z/M/; 
	    $xlabel =~ s/Ohm/Ohm\/s/;
	    $ylabel =~ s/Ohm/Ohm\/s/; 
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    $xlabel =~ s/Z/C\*/;
	    $ylabel =~ s/\-Z/C\*/; 
	    $xlabel =~ s/Ohm/F/;
	    $ylabel =~ s/Ohm/F/; 
	}
	push @l,"set ylabel '".$ylabel."'";
	push @l,"set xlabel '".$xlabel."'";
	push @l,'#OUTPUT#';
	push @l,$self->adv_opt();
	my $lt = 3;
	my $plotstr;
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    push @l,'G(x,y) = x/(x*x+y*y)';
	    push @l,'B(x,y) = -y/(x*x+y*y)';
	    $plotstr = "plot '$filename' u (sf(G(\$2,\$3))):(sf(B(\$2,\$3))) w p,'' u (real(sf(1/f(\$1)))):(imag(sf(1/f(\$1)))) w l lt 2";
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * x';
	    push @l,'Z(x,y) = {1,0} * x + {0,1} * y';
	    $plotstr = "plot '$filename' u (sf(real(Z(\$2,\$3)*M(\$1)))):(sf(imag(Z(\$2,\$3)*M(\$1)))) w p,'' u (real(sf(f(\$1)*M(\$1)))):(imag(sf(f(\$1)*M(\$1)))) w l lt 2";
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * x';
	    push @l,'Z(x,y) = {1,0} * x + {0,1} * y';
	    $plotstr = "plot '$filename' u (sf(real(1/(Z(\$2,\$3)*M(\$1))))):(sf(-imag(1/(Z(\$2,\$3)*M(\$1))))) w p,'' u (real(sf(1/(f(\$1)*M(\$1))))):(-imag(sf(1/(f(\$1)*M(\$1))))) w l lt 2";
	}
	else
	{
	    $plotstr = "plot '$filename' u (sf(\$2)):(sf(-\$3)) w p,'' u (real(sf(f(\$1)))):(-imag(sf(f(\$1)))) w l lt 1";
	    if ($self->{'_show_arcs'})
	    {
		my @list = ();
		my %hash = $self->dev_sort_fmax();
		foreach my $dev (sort { $hash{$b} <=> $hash{$a} } (keys %hash))
		{
		    my $fn = $self->{'_devices'}{$dev}->function_name();
		    $fn =~ s/\(x\)//;
		    my $offsetx = 0;
		    foreach my $d (@list)
		    {
			$offsetx += Re($self->{'_devices'}{$d}->F(1e-7));
		    }
		  OUTER:
		    foreach my $key (keys %{$self->{'_devices'}})
		    {
			foreach (@list)
			{
			    next OUTER if ($_ eq $key);
			}
			$offsetx += Re($self->{'_devices'}{$key}->F(1e100));
		    }
		    my $diff = Re($self->{'_devices'}{$dev}->F(1e-100) - $self->{'_devices'}{$dev}->F(1e100));
		    $offsetx += $diff if ($diff < 0);
		    $plotstr .= ",'' u (sf($offsetx + real($fn(\$1)))):(sf(-imag($fn(\$1)))) w l lt ".$lt++;
		    push @list,$dev;
		}  
	    }
	}
	push @l,$plotstr;
    }
    return join("\n",@l);
}



sub print_plot_error
{
    my $self = shift;
    my $filename = $self->file;
    if (@_)
    {
	$filename = shift;
    }
    unless ($filename)
    {
	$filename = $self->save() if (@{$self->{'_data'}});
    }
    die "No filename specified!" unless ($filename);
    my @l = $self->print_plot_common();
    push @l,'# End Model #';
    push @l,"set ylabel 'Error (%)'";
    push @l,'set logscale x';
    push @l,'set format x "10^{%L}"';
    push @l,"set xlabel 'f(Hz)'";
    push @l,'#OUTPUT#';
    push @l,"plot '$filename' u 1:((\$2 - real(f(\$1)))*100/abs(f(\$1))) w linespoints tit 'Real','' u 1:((\$3 - imag(f(\$1)))*100/abs(f(\$1))) w linespoints tit 'Imag'";
    return join("\n",@l);    
}

sub print_imp_sim($$)
{
    my $self = shift;
    my $minf = shift;
    my $maxf = shift;
    return $self->print_simulation($minf,$maxf);
}

sub print_bode_sim($$)
{
    my $self = shift;
    my $minf = shift;
    my $maxf = shift;
    return $self->print_simulation($minf,$maxf,1);
}

sub print_simulation
{
    my $self = shift;
    my $minf = shift;
    my $maxf = shift;
    $minf = 1 unless (defined($minf));
    $maxf = 10000 unless (defined($maxf));
    die "Illegal range $minf - $maxf" unless ($minf && $minf>0 && $maxf && $maxf>0 && $minf < $maxf);
    my $mode = 0;
    $mode = 1 if (@_);
    my $min = log($minf)/log(10);
    my $max = log($maxf)/log(10);
    my $labelnr = 0;
    my @l = $self->print_plot_common();
    push @l,$self->adv_opt();
    push @l,'# End Model #';
    for (my $x = $min; $x <= $max; $x++)
    {
	my $val = $self->F(10**$x);
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    if ($val != 0)
	    {
		push @l,'set label '.(++$labelnr).' "-- '.POSIX::floor($x+0.5).'" at sf('.Re(1/$val).'), sf('.(Im(1/$val)).'), 0 left rotate' unless ($mode || !$self->plot_label(Re($val),Im($val)));	    
	    }
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    $val = $val * cplx(0,1) * 2 * pi * 10**$x;
	    push @l,'set label '.(++$labelnr).' "-- '.POSIX::floor($x+0.5).'" at sf('.Re($val).'), sf('.(Im($val)).'), 0 left rotate' unless ($mode || !$self->plot_label(Re($val),Im($val)));
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    my $z = $val * cplx(0,1) * 2 * pi * 10**$x;
	    if ($z != 0)
	    {
		$val = 1/$z;
		push @l,'set label '.(++$labelnr).' "-- '.POSIX::floor($x+0.5).'" at sf('.Re($val).'), sf('.(-Im($val)).'), 0 left rotate' unless ($mode || !$self->plot_label(Re($val),Im($val)));	    
	    }
	}
	else
	{
	    push @l,'set label '.(++$labelnr).' "-- '.POSIX::floor($x+0.5).'" at sf('.Re($val).'), sf('.(-Im($val)).'), 0 left rotate' unless ($mode || !$self->plot_label(Re($val),Im($val)));	    
	}
    }
    my $title = $self->model_text;
    push @l,'# End Labels #';
    push @l,'g(x) = f(10**x)';
    push @l,'set parametric';
    push @l,"set trange [$min:$max]";
    if ($mode)
    {
	my $ymin = 1e100;
	if ($self->{'_show_arcs'})
	{
	    for (my $t = $min; $t < $max; $t+= 0.05)
	    {
		my $val = $self->F(10**$t);
		if ($self->{'_plot_admittance'} eq 'Admittance')
		{
		    $val = 1/$val;
		}
		elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
		{
		    $val = $val * cplx(0,1) * 2 * pi * 10**$t;
		}
		elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
		{
		    $val = 1/($val * cplx(0,1) * 2 * pi * 10**$t);
		}		
		my $x = abs(Re($val));
		my $y = abs(Im($val));
		$ymin = $x if ($x < $ymin);
		$ymin = $y if ($y < $ymin);
	    }
	    $ymin = 10 ** POSIX::floor(log($ymin)/log(10) - 1);
	    push @l,"set yrange [$ymin:]";
	}
	my $ylabel = $self->{'_ylabel_bode'};
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    $ylabel =~ s/-Z/Y/; 
	    $ylabel =~ s/Ohm/S/; 
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    $ylabel =~ s/-Z/M/; 
	    $ylabel =~ s/Ohm/Ohm\/s/; 	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    $ylabel =~ s/-Z/C\*/; 
	    $ylabel =~ s/Ohm/F/; 	    	    
	}
	push @l,"set ylabel '".$ylabel."'";
	push @l,"set xlabel 'f(Hz)'";
	push @l,"set logscale x";
	push @l,"set logscale y";
        push @l,'set format x "10^{%L}"';
        push @l,'set format y "10^{%L}"';
	push @l,"p(x) = x>0?x:1/0";
	push @l,"n(x) = x<0?-x:1/0";
	push @l,'#OUTPUT#';
	push @l,$self->adv_opt();
	if ($title =~ /Z = 0/)
	{
	    push @l,'set yrange [1:10]';
	    push @l,"set xrange [$minf:$maxf]";
	}
	my $plotstr = undef;
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    $plotstr = "plot 10**t,sf(p(real(1/g(t)))) w l lt 1 linewidth 3 tit '$title (Real+)',10**t,sf(n(real(1/g(t)))) w l lt 2 linewidth 3 tit '$title (Real-)',10**t,sf(p(imag(1/g(t)))) w l lt 4 linewidth 3 tit '$title (Imag+)',10**t,sf(n(imag(1/g(t)))) w l lt 3 linewidth 3 tit '$title (Imag-)'";
	    $plotstr = "plot 10**t,sf(p(imag(1/g(t)))) w l lt 4 linewidth 3 tit '$title (Imag+)',10**t,sf(n(imag(1/g(t)))) w l lt 3 linewidth 3 tit '$title (Imag-)'" if ($self->{'_bode_mode'});
	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * 10**x';
	    $plotstr = "plot 10**t,sf(p(real(g(t)*M(t)))) w l lt 1 linewidth 3 tit '$title (Real+)',10**t,sf(n(real(g(t)*M(t)))) w l lt 2 linewidth 3 tit '$title (Real-)',10**t,sf(p(imag(g(t)*M(t)))) w l lt 4 linewidth 3 tit '$title (Imag+)',10**t,sf(n(imag(g(t)*M(t)))) w l lt 3 linewidth 3 tit '$title (Imag-)'";
	    $plotstr = "plot 10**t,sf(p(imag(g(t)*M(t)))) w l lt 4 linewidth 3 tit '$title (Imag+)',10**t,sf(n(imag(g(t)*M(t)))) w l lt 3 linewidth 3 tit '$title (Imag-)'" if ($self->{'_bode_mode'});	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * 10**x';
	    $plotstr = "plot 10**t,sf(p(real(1/(g(t)*M(t))))) w l lt 1 linewidth 3 tit '$title (Real+)',10**t,sf(n(real(1/(g(t)*M(t))))) w l lt 2 linewidth 3 tit '$title (Real-)',10**t,sf(p(imag(1/(g(t)*M(t))))) w l lt 4 linewidth 3 tit '$title (Imag+)',10**t,sf(n(imag(1/(g(t)*M(t))))) w l lt 3 linewidth 3 tit '$title (Imag-)'";
	    $plotstr = "plot 10**t,sf(p(imag(1/(g(t)*M(t))))) w l lt 4 linewidth 3 tit '$title (Imag+)',10**t,sf(n(imag(1/(g(t)*M(t))))) w l lt 3 linewidth 3 tit '$title (Imag-)'" if ($self->{'_bode_mode'});	    	    
	}
	else
	{
	    $plotstr = "plot 10**t,sf(p(real(g(t)))) w l lt 1 linewidth 3 tit '$title (Real+)',10**t,sf(n(real(g(t)))) w l lt 2 linewidth 3 tit '$title (Real-)',10**t,sf(p(imag(g(t)))) w l lt 4 linewidth 3 tit '$title (Imag+)',10**t,sf(n(imag(g(t)))) w l lt 3 linewidth 3 tit '$title (Imag-)'";
	    $plotstr = "plot 10**t,sf(p(imag(g(t)))) w l lt 4 linewidth 3 tit '$title (Imag+)',10**t,sf(n(imag(g(t)))) w l lt 3 linewidth 3 tit '$title (Imag-)'" if ($self->{'_bode_mode'});	    
	}
	if ($self->{'_show_arcs'})
	{
	    my $lt = 5;
	    my %hash = $self->dev_sort_fmax();
	    foreach my $dev (sort { $hash{$b} <=> $hash{$a} } (keys %hash))
	    {
		my $fn = $self->{'_devices'}{$dev}->function_name();
		my $name = $self->{'_devices'}{$dev}->name();
		$fn =~ s/\(x\)//;
		if ($self->{'_plot_admittance'} eq 'Admittance')
		{
		    $plotstr .= ", 10**t,sf(abs(imag(1/$fn(10**t)))) w l lt ".$lt++."tit '$name'";
		}
		elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
		{
		    $plotstr .= ", 10**t,sf(abs(imag($fn(10**t)*M(t)))) w l lt ".$lt++."tit '$name'";
		}
		elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
		{
		    $plotstr .= ", 10**t,sf(abs(imag(1/($fn(10**t)*M(t))))) w l lt ".$lt++."tit '$name'";
		}
		else
		{
		    $plotstr .= ", 10**t,sf(abs(imag($fn(10**t)))) w l lt ".$lt++."tit '$name'";
		}
	    }
	}
	push @l,$plotstr;
    }
    else
    {
	push @l,'set size ratio -1';
	my $xlabel = $self->{'_xlabel'};
	my $ylabel = $self->{'_ylabel'};
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    $xlabel =~ s/Z/Y/;
	    $ylabel =~ s/\-Z/Y/; 
	    $xlabel =~ s/Ohm/S/;
	    $ylabel =~ s/Ohm/S/; 
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    $xlabel =~ s/Z/M/;
	    $ylabel =~ s/\-Z/M/; 
	    $xlabel =~ s/Ohm/Ohm\/s/;
	    $ylabel =~ s/Ohm/Ohm\/s/; 
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    $xlabel =~ s/Z/C\*/;
	    $ylabel =~ s/\-Z/C\*/; 
	    $xlabel =~ s/Ohm/F/;
	    $ylabel =~ s/Ohm/F/; 
	}
	push @l,"set ylabel '".$ylabel."'";
	push @l,"set xlabel '".$xlabel."'";
	push @l,'#OUTPUT#';
	push @l,$self->adv_opt();
	my $lt = 5;
	my $plotstr;
	if ($self->{'_plot_admittance'} eq 'Admittance')
	{
	    $plotstr = "plot real(sf(1/g(t))),imag(sf(1/g(t))) w l linewidth 3 tit '$title'";	    
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex modulus')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * 10**x';
	    $plotstr = "plot real(sf(g(t)*M(t))),imag(sf(g(t)*M(t))) w l linewidth 3 tit '$title'";
#	    if ($self->{'_show_arcs'})
#	    {
#		my @list = ();
#		my %hash = $self->dev_sort_fmax();
#		foreach my $dev (reverse sort { $hash{$b} <=> $hash{$a} } (keys %hash))
#		{
#		    my $fn = $self->{'_devices'}{$dev}->function_name();
#		    my $name = $self->{'_devices'}{$dev}->name();
#		    $fn =~ s/\(x\)//;
#		    my $offsetx = 0;
#		    foreach my $d (@list)
#		    {
#			$offsetx += Re($self->{'_devices'}{$d}->F(1e100) * cplx(0,1) * 2 * pi * 1e100);
#		    }
#		    $plotstr .= ", sf($offsetx + real($fn(10**t)*M(t))),imag(sf($fn(10**t)*M(t))) w l lt ".$lt++."tit '$name'";
#		    push @list,$dev;
#		}  
#	    }
	}
	elsif ($self->{'_plot_admittance'} eq 'Complex capacitance')
	{
	    push @l,'M(x) = {0,1} * 2 * pi * 10**x';
	    $plotstr = "plot real(sf(1/(g(t)*M(t)))),-imag(sf(1/(g(t)*M(t)))) w l linewidth 3 tit '$title'";
	}
	else
	{
	    $plotstr = "plot real(sf(g(t))),-imag(sf(g(t))) w l linewidth 3 tit '$title'";	    
	    if ($self->{'_show_arcs'})
	    {
		my @list = ();
		my %hash = $self->dev_sort_fmax();
		foreach my $dev (sort { $hash{$b} <=> $hash{$a} } (keys %hash))
		{
		    my $fn = $self->{'_devices'}{$dev}->function_name();
		    my $name = $self->{'_devices'}{$dev}->name();
		    $fn =~ s/\(x\)//;
		    my $offsetx = 0;
		    foreach my $d (@list)
		    {
			$offsetx += Re($self->{'_devices'}{$d}->F(1e-100));
		    }
		  OUTER:
		    foreach my $key (keys %{$self->{'_devices'}})
		    {
			foreach (@list)
			{
			    next OUTER if ($_ eq $key);
			}
			$offsetx += Re($self->{'_devices'}{$key}->F(1e100));
		    }
		    my $diff = Re($self->{'_devices'}{$dev}->F(1e-100) - $self->{'_devices'}{$dev}->F(1e100));
		    $offsetx += $diff if ($diff < 0);
		    $plotstr .= ", sf($offsetx + real($fn(10**t))),-imag(sf($fn(10**t))) w l lt ".$lt++."tit '$name'";
		    push @list,$dev;
		}  
	    }
	}
	push @l,$plotstr;
    }
    return join("\n",@l);
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    my $res = cplx(0,0);
    foreach (keys %{$self->{'_devices'}})
    {
	$res += $self->{'_devices'}{$_}->F($f);
    }
    return $res;
}

sub subset($$)
{
    my $self = shift;
    my $fmin = shift;
    my $fmax = shift;
    die "Undefined frequency range\n" unless ($fmax && $fmin);
    die "Illegal frequency range $fmin:$fmax\n" unless ($fmax > $fmin);
    my @res = ();
    foreach (@{$self->{'_data'}})
    {
	push @res,join(' ',@{$_}) if ($$_[0] >= $fmin && $$_[0] <= $fmax);
    }
    return join("\n",@res);
}

sub delete_point($)
{
    my $self = shift;
    my $index = shift;
    die "Undefined index or index not a number!" unless (defined($index) && $index =~ /\d/);
    my @data = ();
    @data = split(/\n/,$self->data());
    die "Index out of bounds!" unless ($index > -1 && $index < @data);
    my @newdata = ();
    for(my $x = 0; $x < @data; $x++)
    {
	next if ($x == $index);
	push @newdata,$data[$x];
    }
    return $self->data(join("\n",@newdata));    
}

sub data
{
    my $self = shift;
    if (@_)
    {
	my $str = shift;
	$self->{'_data'} = [];
	return undef unless ($str);
	my @list = split(/\n/,$str);
	my %h = ();
	foreach my $s (@list)
	{
	    $s =~ s/^[\s\t]*//;
	    my @l = split(/[\s\t]+/,$s);
	    if (@l == 3)
	    {
		foreach my $n (@l)
		{
		    $n += 0;
		}
		### Only include lines with valid frequency ###
		$h{$l[0]} = \@l unless ($l[0] == 0); 
	    }
	}
	foreach (reverse(sort {$a<=>$b} (keys %h)))
	{
	    push @{$self->{'_data'}},$h{$_}; 
	}
    }
    my @res = ();
    foreach (@{$self->{'_data'}})
    {
	push @res,join(' ',@{$_});
    }
    return join("\n",@res);
}


sub save()
{
    my $self = shift;
    $self =~ /\((\w+)\)/;
    my $filename = '/tmp/Impedance_'.$$.'_'.$1.'.dat';
    push @{$self->{'_unlinkfiles'}},$filename;
    $self->{'_file'} = SemaforeFile->new($filename);
    $self->{'_file'}->writeline($self->data);
    $self->{'_filename'} = $self->{'_file'}->filename();
    return $self->{'_filename'};
}

sub file()
{
    my $self = shift;
    return $self->{'_filename'};
}

sub subtract_model()
{
    my $self = shift;
    return undef unless (@{$self->{'_data'}});
    my @res = () ;
    foreach (@{$self->{'_data'}})
    {
	my $f = $$_[0];
	my $z = cplx($$_[1],$$_[2]);
	$z = $z - $self->F($f);
	push @res,sprintf("%.6g %.6g %.6g",$f,Re($z),Im($z));
    }
    return join("\n",@res);    
}

sub get_error()
{
    my $self = shift;
    return (undef,undef) unless (@{$self->{'_data'}});
    my @res = () ;
    my $max = 0;
    my $n = 0;
    my $sum = 0;
    foreach (@{$self->{'_data'}})
    {
	my $f = $$_[0];
	my $z = cplx($$_[1],$$_[2]);
	my $zm = $self->F($f);
	my $e = $z - $zm;
	foreach ($e->Re(),$e->Im())
	{
	    next unless ($zm->abs());
	    my $x = abs($_) * 100 / $zm->abs();
	    push @res, $x;
	    $sum += $x;
	    $n++;
	    $max = $x if ($x > $max);
	}
    }
    my $mean = $sum / $n;
    return ($mean,$max);
}

sub DESTROY()
{
    my $self = shift;
    unlink $self->{'_filename'} if ($self->{'_filename'} && -e $self->{'_filename'});
    unlink $self->{'_fitset'} if ($self->{'_fitset'} && -e $self->{'_fitset'});
    foreach my $fn (@{$self->{'_unlinkfiles'}})
    {
	unlink $fn if ($fn && -e $fn);
    }
}


return 1;
