package Impedance::RQ;
#$Id: RQ.pm,v 1.22 2015/10/21 12:04:33 sofc Exp $

######################################################################
#
#   Package Impedance::RQ
#
# This module implements Impedance::Base for R-C and R-L parallel connections 
# as well as the R-Q parallel connection. (The Q beeing a constant phase 
# element).
#
# To obtain an instance, call one of the constructors:
#
# $id = Impedance::RC->new($id);
#
# $id = Impedance::RL->new($id);
#
# $id = Impedance::RQ->new($id);
#
# The module defines no additional member functions.
#
###################################################################

use strict;
use warnings;
use Impedance::Base;
use Math::Complex;
our @ISA = qw(Impedance::Base);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'RQ';
    $self->{'_tags'} = ['R','Y','n'];
    $self->{'_helptags'} = ['Y','R','n'];
    $self->{'_values'}{'Y'} = 1e-4; ### Default 100yF ###
    $self->{'_values'}{'n'} = 0.8; ### Default 0.8 ###
    $self->{'_desc'}{'name'} = 'Parallel connection of resistor and constant phase element';
    $self->{'_desc'}{'R'} = 'Resistance of the resistor';
    $self->{'_desc'}{'Y'} = "'Capacitance' of the constant phase element in F*s^(n-1)";
    $self->{'_desc'}{'n'} = "Phase angle in '1/2 pi' radians (1 mean 90 degrees)";
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    return (1/(2 * pi())) * (1/($self->value('Y')*$self->value('R'))) ** (1/$self->value('n'));
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("R_$n = ".$self->value('R'),"fr_$n = 1.0","Y_$n = ".$self->value('Y'),"fy_$n = 1.0","n_$n = ".$self->value('n'));
    push @l,'z_'.$n.'(x)=1.0/(1.0/(R_'.$n.'*fr_'.$n.')+Y_'.$n.'*fy_'.$n.'*({0,1}*2*pi*x)**n_'.$n.')';
    return join("\n",@l)."\n";
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return cplx(1,0) / (1 / $self->value('R') + $self->value('Y') * (cplx(0,1) * 2 * pi() * $f )**$self->value('n'));
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fr_$n","fy_$n","n_$n");
}

sub save()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    push @l,"Final_R_$n = R_$n * fr_$n";
    push @l,"Final_Y_$n = Y_$n * fy_$n";
    push @l,"Final_n_$n = n_$n";
    push @l,"Final_Fmax_$n = (1/(2*pi))*(1/(Final_R_$n * Final_Y_$n))**(1/n_$n)";
    push @l,"Final_Cps_$n = (Final_R_$n * Final_Y_$n)**(1/n_$n) / Final_R_$n";
    return join("\n",@l)."\n";
}

sub C_start($$$)
{
    ### Returns a guess at the 'capacitance' of the Q element ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    return undef unless ($minf && $maxf && $sumf);
    my @l = split(/[\s\t]+/,$minf);
    my $min = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$maxf);
    my $max = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$sumf);
    my $sum = cplx($l[1],$l[2]);
    my $r = Re($min) - Re($max);
    my $n = -Im($sum) * 2 / $r;
    my $x = $l[0];  ### Summit frequency in Hz   ##
    return 1/($r*(2*$self->pi()*$x)**$n);
}

sub n_start($$$)
{
    ### Returns a guess at the 'n' of the Q element ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    return undef unless ($minf && $maxf && $sumf);
    my @l = split(/[\s\t]+/,$minf);
    my $min = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$maxf);
    my $max = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$sumf);
    my $sum = cplx($l[1],$l[2]);
    my $r = Re($min) - Re($max);
    return -2 * Im($sum) / $r;
}

sub R_start($$$)
{
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    return undef unless ($minf && $maxf && $sumf);
    my @l = split(/[\s\t]+/,$minf);
    my $min = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$maxf);
    my $max = cplx($l[1],$l[2]);
    return Re($min) - Re($max);
}

sub helpfunction
{
    #### Usage: $id->helpfunction($tag,$minf,$sumf,$maxf);
    #### where the frequency variables is a line of the form:
    #### frequency real_value imaginary_value
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return "Select minimum frequency, summit frequency and maximum frequency for arc" unless (@_);
    return $self->C_start(@_) if ($tag eq 'Y');
    return $self->R_start(@_) if ($tag eq 'R');
    return $self->n_start(@_) if ($tag eq 'n');
    return undef;
}

package Impedance::RC;
use Math::Complex;
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'RC';
    $self->{'_tags'} = ['R','C'];
    $self->{'_helptags'} = ['C','R'];
    $self->{'_values'}{'C'} = 1e-4; ### Default 100yF ###
    $self->{'_desc'}{'name'} = 'Parallel connection of resistor and capacitor';
    $self->{'_desc'}{'R'} = 'Resistance of the resistor';
    $self->{'_desc'}{'C'} = "Capacitance of the capacitor";
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    return (1/(2 * pi())) / ($self->value('C') * $self->value('R'));
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("R_$n = ".$self->value('R'),"fr_$n = 1.0","C_$n = ".$self->value('C'),"fc_$n = 1.0");
    push @l,'z_'.$n.'(x)=1.0/(1.0/(R_'.$n.'*fr_'.$n.')+C_'.$n.'*fc_'.$n.'*({0,1}*2*pi*x))';
    return join("\n",@l)."\n";
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return cplx(1,0) / (1 / $self->value('R') + $self->value('C') * (cplx(0,1) * 2 * pi() * $f ));
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fr_$n","fc_$n");
}

sub C_start($$)
{
    ### Returns a guess at the capacitance ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    return undef unless ($minf && $maxf && $sumf);
    my @l = split(/[\s\t]+/,$minf);
    my $min = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$maxf);
    my $max = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$sumf);
    my $x = $l[0];  ### Summit frequency in Hz   ##
    my $r = Re($min) - Re($max);
    return 1/(2*$self->pi()*$r*$x);
}

sub R_start($$$)
{
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    return undef unless ($minf && $maxf && $sumf);
    my @l = split(/[\s\t]+/,$minf);
    my $min = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$maxf);
    my $max = cplx($l[1],$l[2]);
    return Re($min) - Re($max);
}

sub save()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    push @l,"Final_R_$n = R_$n * fr_$n";
    push @l,"Final_C_$n = C_$n * fc_$n";
    push @l,"Final_Fmax_$n = (1/(2*pi))/(Final_R_$n * Final_C_$n) ";
    return join("\n",@l)."\n";
}

sub helpfunction()
{
    #### Usage: $id->helpfunction($tag,$minf,$sumf,$maxf);
    #### where the frequency variables is a line of the form:
    #### frequency real_value imaginary_value
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return "Select minimum frequency, summit frequency and maximum frequency for arc" unless (@_);
    return $self->C_start(@_) if ($tag eq 'C');
    return $self->R_start(@_) if ($tag eq 'R');
    return undef;
}

package Impedance::RL;
use Math::Complex;
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'RL';
    $self->{'_tags'} = ['R','L'];
    $self->{'_helptags'} = ['L','R'];
    $self->{'_values'}{'L'} = 1e-7; ### Default 100 nH ###
    $self->{'_desc'}{'name'} = 'Parallel connection of resistor and inductor';
    $self->{'_desc'}{'R'} = 'Resistance of the resistor';
    $self->{'_desc'}{'L'} = 'Inductance of the inductor';
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    return $self->value('R') / (2 * pi() * $self->value('L'));
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("R_$n = ".$self->value('R'),"fr_$n = 1.0","L_$n = ".$self->value('L'),"fl_$n = 1.0");
    push @l,'z_'.$n.'(x)=1.0/(1.0/(R_'.$n.'*fr_'.$n.') + 1.0/(L_'.$n.'*fl_'.$n.'*({0,1}*2*pi*x)))';
    return join("\n",@l)."\n";
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return cplx(1,0) / (1 / $self->value('R') +  1 /($self->value('L') * (cplx(0,1) * 2 * pi() * $f )));
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fr_$n","fl_$n");
}

sub L_start($$)
{
    ### Returns a guess at the inductance ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    return undef unless ($minf && $maxf && $sumf);
    my $r = $self->R_start($minf,$sumf,$maxf);
    my @l = split(/[\s\t]+/,$sumf);
    my $x = $l[0];  ### Summit frequency in Hz   ##
    return $r/(2*$self->pi()*$x);
}

sub R_start($$$)
{
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    return undef unless ($minf && $maxf && $sumf);
    my @l = split(/[\s\t]+/,$minf);
    my $min = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$maxf);
    my $max = cplx($l[1],$l[2]);
    return abs(Re($max) - Re($min));
}

sub save()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    push @l,"Final_R_$n = R_$n * fr_$n";
    push @l,"Final_L_$n = L_$n * fl_$n";
    push @l,"Final_Fmax_$n = Final_R_$n / ( 2 * pi * Final_L_$n)";
    return join("\n",@l)."\n";
}

sub helpfunction()
{
    #### Usage: $id->helpfunction($tag,$minf,$sumf,$maxf);
    #### where the frequency variables is a line of the form:
    #### frequency real_value imaginary_value
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return "Select minimum frequency, summit frequency and maximum frequency for arc" unless (@_);
    return $self->L_start(@_) if ($tag eq 'L');
    return $self->R_start(@_) if ($tag eq 'R');
    return undef;
}

return 1;

