package Impedance::W;
#$Id: W.pm,v 1.40 2023/02/28 10:01:33 sqko Exp $
#####################################################################
#
# Package Impedance::W;
#
# This module implements Impedance::Base for diffusion type elements
# (W, O, Od, G, Gd and T).
# W is the Warburg element, O is the Finite length Warburg element,
# Od is a depressed / flattened finite length warburg element,
# G is the Gerisher element, Gd is a depressed / flattened Gerisher element, 
# T is the 'Bounded Warburg' element and dL is the 'de Levie' element.
#
# To obtain an instance call one of the constructors:
#
# $id = Impedance::W->new($nr);
#
# $id = Impedance::O->new($nr);
#
# $id = Impedance::Od->new($nr);
#
# $id = Impedance::G->new($nr);
#
# $id = Impedance::Gd->new($nr);
#
# $id = Impedance::T->new($nr);
#
# $id = Impedance::dL->new($nr);
#
# The module defines no additional member functions.
#
#####################################################################
use strict;
use warnings;
use Impedance::Base;
use Math::Complex;
our @ISA = qw(Impedance::Base);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'W';
    $self->{'_tags'} = ['Y'];
    $self->{'_desc'}{'name'} = "Warburg diffusion element (infinite linear diffusion)";
    $self->{'_desc'}{'Y'} = "Admittance in S*s^(1/2)";
    $self->{'_f_imax'} = 1e-100;
    bless $self,$class;
    return $self;
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("Y_$n = ".$self->value('Y'),"fy_$n = 1.0");
    push @l,'z_'.$n.'(x)={1,-1} / (2**0.5*Y_'.$n.'*fy_'.$n.'*(pi*2*x)**0.5)';
    return join("\n",@l)."\n";
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return cplx(1,-1) / (2**0.5 * $self->value('Y') * (2 * pi() * $f)**0.5);
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fy_$n");
}

package Impedance::O;
use Math::Complex;
### Finite length warburg impedance ###
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'O';
    $self->{'_tags'} = ['Zo','B'];
    $self->{'_helptags'} = ['B','Zo'];
    $self->{'_desc'}{'name'} = "Warburg diffusion element of finite length";
    $self->{'_desc'}{'Zo'} = "DC-Impedance of element in Ohm";
    $self->{'_desc'}{'B'} = "'Time constant' in s^0.5";
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    return 0.4 / ($self->value('B') ** 2);
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("Zo_$n = ".$self->value('Zo'),"fzo_$n = 1.0","B_$n = ".$self->value('B'),"fb_$n = 1.0");
    push @l,'tnh(x) = (abs(x)>100 ? {1,0} : tanh(x))';
    push @l,'z_'.$n.'(x) = Zo_'.$n.' * fzo_'.$n.' * tnh( (B_'.$n.' * fb_'.$n.' * {0,1}*2*pi*x)**0.5 ) / ((B_'.$n.' * fb_'.$n.' * {0,1}*2*pi*x)**0.5)';
    return join("\n",@l)."\n";
}


sub save()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    push @l,"Final_B_$n = B_$n * fb_$n";
    push @l,"Final_Zo_$n = Zo_$n * fzo_$n";
    push @l,"Final_Fmax_$n = 0.4 / (B_$n * B_$n)";
    return join("\n",@l)."\n";
}

sub my_tanh($)
{
    my $self = shift;
    my $arg = shift;
    if ($arg > 100)
    {
	return cplx(1,0);
    }
    if ($arg < 1e-10)
    {
	return $arg;
    }
    else
    {
	return tanh($arg);
    }
}


sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    my $t = $self->value('Zo') * $self->my_tanh( ($self->value('B') * cplx(0,1) * 2 * pi() * $f)**0.5);
    my $n = ($self->value('B') * cplx(0,1) * 2 * pi() * $f)**0.5;
    return $t/$n;
}

sub B_start($$$)
{
    ### Returns a guess at the 'B' parameter of the elemenet ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $x = $l[0];  ### Summit frequency in Hz   ##
    return sqrt(0.4/$x);
}

sub Zo_start($$$)
{
    ### Returns a guess at the 'capacitance' of the element ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $B = sqrt(0.4/$l[0]); ### gest guess at B parameter                   ##
    my @m = split(/[\s\t]+/,$maxf);
    @l = split(/[\s\t]+/,$minf);
    my $z_max = $l[1] - $m[1];  ### max impedance (intercept with x for element ##
    return $z_max;
}

sub helpfunction
{
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return "Select minimum frequency, sumit frequency and maximum frequency for arc" unless (@_);
    return $self->Zo_start(@_) if ($tag eq 'Zo');
    return $self->B_start(@_) if ($tag eq 'B');
    return undef;
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fzo_$n","fb_$n");
}

package Impedance::Od;
use Math::Complex;
### Depressed finite length warburg impedance ###
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'Od';
    $self->{'_tags'} = ['Zo','B','n'];
    $self->{'_helptags'} = ['n','B','Zo'];
    $self->{'_values'}{'n'} = 0.4;
    $self->{'_desc'}{'name'} = "Depressed warburg diffusion element of finite length";
    $self->{'_desc'}{'Zo'} = "DC-Impedance of element in Ohm";
    $self->{'_desc'}{'B'} = "'Time constant' in s^n";
    $self->{'_desc'}{'n'} = "Exponent";
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    return 0.8 * $self->value('n') / ($self->value('B') ** 2);
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("Zo_$n = ".$self->value('Zo'),"fzo_$n = 1.0","B_$n = ".$self->value('B'),"fb_$n = 1.0","n_$n = ".$self->value('n'),"fn_$n = 1.0");
    push @l,'tnh(x) = (abs(x)>100 ? {1,0} : tanh(x))';
    push @l,'z_'.$n.'(x) = Zo_'.$n.' * fzo_'.$n.' * tnh( (B_'.$n.' * fb_'.$n.' * {0,1}*2*pi*x)**(fn_'.$n.' * n_'.$n.') ) / ((B_'.$n.' * fb_'.$n.' * {0,1}*2*pi*x)**(fn_'.$n.' * n_'.$n.'))';
    return join("\n",@l)."\n";
}

sub save()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    push @l,"Final_B_$n = B_$n * fb_$n";
    push @l,"Final_Zo_$n = Zo_$n * fzo_$n";
    push @l,"Final_n_$n = n_$n * fn_$n";
    push @l,"Final_Fmax_$n = 0.8 * Final_n_$n / (Final_B_$n * Final_B_$n)";
    return join("\n",@l)."\n";
}

sub my_tanh($)
{
    my $self = shift;
    my $arg = shift;
    if ($arg > 100)
    {
	return cplx(1,0);
    }
    if ($arg < 1e-10)
    {
	return $arg;
    }
    else
    {
	return tanh($arg);
    }
}


sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return $self->value('Zo') * $self->my_tanh( ($self->value('B') * cplx(0,1) * 2 * pi() * $f)**$self->value('n')) / ($self->value('B') * cplx(0,1) * 2 * pi() * $f)**$self->value('n');
}

sub n_start($$$)
{
    ### Returns a guess at the 'n' parameter of the elemenet              ###
    ### Determined empirically and is accurately enough for start guesses ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$minf);
    my $min = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$maxf);
    my $max = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$sumf);
    my $sum = cplx($l[1],$l[2]);
    my $r = Re($max) - Re($min);
    my $i = Im($sum) - Im($max);
    my $n = 1.5 * $i / $r;
    return $n;
}

sub B_start($$$)
{
    ### Returns a guess at the 'B' parameter of the elemenet ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $x = $l[0];  ### Summit frequency in Hz   ##
    return sqrt(0.8 * $self->n_start($minf,$sumf,$maxf)/$x);
}

sub Zo_start($$$)
{
    ### Returns a guess at the 'capacitance' of the element ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $B = $self->B_start($minf,$sumf,$maxf);
    my @m = split(/[\s\t]+/,$maxf);
    @l = split(/[\s\t]+/,$minf);
    my $z_max = $l[1] - $m[1];  ### max impedance (intercept with x for element ##
    return $z_max;
}


sub helpfunction
{
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return "Select minimum frequency, sumit frequency and maximum frequency for arc" unless (@_);
    return $self->Zo_start(@_) if ($tag eq 'Zo');
    return $self->B_start(@_) if ($tag eq 'B');
    return $self->n_start(@_) if ($tag eq 'n');
    return undef;
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fzo_$n","fb_$n","fn_$n");
}

package Impedance::T;
use Math::Complex;
### Finite capacity warburg impedance ###
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'T';
    $self->{'_tags'} = ['Zo','B'];
    $self->{'_helptags'} = ['Zo','B'];
    $self->{'_desc'}{'name'} = "Warburg diffusion element of finite capacity";
    $self->{'_desc'}{'Zo'} = "Scale factor in Ohm";
    $self->{'_desc'}{'B'} = "'Time constant' in s^(1/2)";
    $self->{'_B_hash'} = {
	     95 => 0.1,
	     25 => 0.2,
	     10 => 0.3,
	     5 => 0.4,
	     3 => 0.5,
	     2 => 0.6,
	     1.5 => 0.7,
	     1.25 => 0.8,
	     1 => 0.9,
	     0.75 => 0.9,
	     0.6 => 1.1,
	     0.5 => 1.2,
	     0.4 => 1.3,
	     0.3 => 1.5,
	     0.15 => 2,
	     0.1 => 3,
	     0.05 => 5,
	     0.01 => 10,
	     0.002 => 20,
	     };
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    foreach my $f (reverse sort {$a <=> $b} keys %{$self->{'_B_hash'}})
    {
	next if $self->{'_B_hash'}{$f} < $self->value('B');
	return $f;
    }
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("Zo_$n = ".$self->value('Zo'),"fzo_$n = 1.0","B_$n = ".$self->value('B'),"fb_$n = 1.0");
    push @l,'cth(x) = (abs(x)>50 ? {1,0} : cosh(x)/sinh(x))';
    push @l,'z_'.$n.'(x) = Zo_'.$n.' * fzo_'.$n.' *  cth( (B_'.$n.' * fb_'.$n.' * {0,1}*2*pi*x)**0.5 ) / (B_'.$n.' * fb_'.$n.' * {0,1}*2*pi*x)**0.5';
    return join("\n",@l)."\n";
}

sub my_coth($)
{
    my $self = shift;
    my $val = shift;
    if ($val > 50)
    {
	return cplx(1,0);
    }
    else
    {
	return coth($val);
    }
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    my $t = $self->value('Zo') * $self->my_coth( ($self->value('B') * cplx(0,1) * 2 * pi() * $f)**0.5);
    my $n = ($self->value('B') * cplx(0,1) * 2 * pi() * $f)**0.5;
    return $t/$n;
    return $self->value('Zo') * $self->my_coth( ($self->value('B') * cplx(0,1) * 2 * pi() * $f)**0.5) / ($self->value('B') * cplx(0,1) * 2 * pi() * $f)**0.5;
}

sub B_start($$$)
{
    ### Returns a guess at the 'B' parameter of the elemenet ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $x = $l[0];  ### 'kink' frequency in Hz   ##
    foreach my $f (reverse sort {$a <=> $b} keys %{$self->{'_B_hash'}})
    {
	next if $f > $x;
	return $self->{'_B_hash'}{$f};
    }
}

sub Zo_start($$$)
{
    ### Returns a guess at the 'capacitance' of the element ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $B = $self->B_start($minf,$sumf,$maxf);
    my @m = split(/[\s\t]+/,$maxf);
    @l = split(/[\s\t]+/,$minf);
    my $z_max = $l[1] - $m[1];  ### max impedance (intercept with x for element ##
    return 3 * $z_max;
}

sub helpfunction
{
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return "Select minimum frequency, bend frequency and maximum frequency" unless (@_);
    return $self->Zo_start(@_) if ($tag eq 'Zo');
    return $self->B_start(@_) if ($tag eq 'B');
    return undef;
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fzo_$n","fb_$n");
}

package Impedance::Td;
use Math::Complex;
### Depressed finite capacity warburg impedance ###
### Warning, this element may be wrong!         ###
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'Td';
    $self->{'_tags'} = ['Zo','B','n'];
    $self->{'_helptags'} = ['Zo','B','n'];
    $self->{'_values'}{'n'} = 0.4;
    $self->{'_desc'}{'name'} = "Depressed warburg diffusion element of finite capacity";
    $self->{'_desc'}{'Zo'} = "Scale factor in Ohm";
    $self->{'_desc'}{'B'} = "'Time constant' in s^n";
    $self->{'_desc'}{'n'} = "Exponent";
    $self->{'_B_hash'} = {
	     95 => 0.1,
	     25 => 0.2,
	     10 => 0.3,
	     5 => 0.4,
	     3 => 0.5,
	     2 => 0.6,
	     1.5 => 0.7,
	     1.25 => 0.8,
	     1 => 0.9,
	     0.75 => 0.9,
	     0.6 => 1.1,
	     0.5 => 1.2,
	     0.4 => 1.3,
	     0.3 => 1.5,
	     0.15 => 2,
	     0.1 => 3,
	     0.05 => 5,
	     0.01 => 10,
	     0.002 => 20,
	     };
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    foreach my $f (reverse sort {$a <=> $b} keys %{$self->{'_B_hash'}})
    {
	next if $self->{'_B_hash'}{$f} < $self->value('B');
	return $f;
    }
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("Zo_$n = ".$self->value('Zo'),"fzo_$n = 1.0","B_$n = ".$self->value('B'),"fb_$n = 1.0","n_$n = ".$self->value('n'),"fn_$n = 1.0");
    push @l,'cth(x) = (abs(x)>50 ? {1,0} : cosh(x)/sinh(x))';
    push @l,'z_'.$n.'(x)= Zo_'.$n.' * fzo_'.$n.' * cth( (B_'.$n.' * fb_'.$n.' * {0,1}*2*pi*x)**(n_'.$n.' * fn_'.$n.') ) / (B_'.$n.' * fb_'.$n.' * {0,1}*2*pi*x)**(n_'.$n.' * fn_'.$n.')';
    return join("\n",@l)."\n";
}

sub my_coth($)
{
    my $self = shift;
    my $val = shift;
    if ($val > 50)
    {
	return cplx(1,0);
    }
    else
    {
	return cosh($val)/sinh($val);
    }
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return $self->value('Zo') * $self->my_coth( ( $self->value('B') * cplx(0,1) * 2 * pi() * $f)**$self->value('n')) / ($self->value('B') * cplx(0,1) * 2 * pi() * $f)**$self->value('n');
}

sub n_start($$$)
{
    ### Returns a guess at the 'n' parameter of the elemenet              ###
    ### Determined empirically and is accurately enough for start guesses ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$minf);
    my $min = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$maxf);
    my $max = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$sumf);
    my $sum = cplx($l[1],$l[2]);
    my $r = Re($sum) - Re($max);
    my $i = Im($max) - Im($sum);
    my $n = 0.5 * $i / $r;
    return $n;
}

sub B_start($$$)
{
    ### Returns a guess at the 'B' parameter of the elemenet ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $x = $l[0];  ### 'kink' frequency in Hz   ##
    foreach my $f (reverse sort {$a <=> $b} keys %{$self->{'_B_hash'}})
    {
	next if $f > $x;
	return $self->{'_B_hash'}{$f};
    }
}

sub Zo_start($$$)
{
    ### Returns a guess at the 'capacitance' of the element ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $B = $self->B_start($minf,$sumf,$maxf);
    my @m = split(/[\s\t]+/,$maxf);
    @l = split(/[\s\t]+/,$minf);
    my $z_max = $l[1] - $m[1];  ### max impedance (intercept with x for element ##
    return 3 * $z_max;
}

sub helpfunction
{
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return "Select minimum frequency, bend frequency and maximum frequency" unless (@_);
    return $self->Zo_start(@_) if ($tag eq 'Zo');
    return $self->B_start(@_) if ($tag eq 'B');
    return $self->n_start(@_) if ($tag eq 'n');
    return undef;
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fzo_$n","fb_$n","fn_$n");
}


package Impedance::G;
use Math::Complex;
### Gerisher impedance ###
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'G';
    $self->{'_tags'} = ['Y','k'];
    $self->{'_helptags'} = ['Y','k'];
    $self->{'_desc'}{'name'} = "Gerisher diffusion element";
    $self->{'_desc'}{'Y'} = "'Admittance' of element in S*s^(1/2)";
    $self->{'_desc'}{'k'} = "'Rate constant' in s^(-1)";
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    return $self->value('k')/4;
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("Y_$n = ".$self->value('Y'),"fy_$n = 1.0","k_$n = ".$self->value('k'),"fk_$n = 1.0");
    push @l,'z_'.$n.'(x)=1.0 / ( Y_'.$n.' * fy_'.$n.' * (k_'.$n.' * fk_'.$n.' + {0,1}*2*pi*x)**0.5)';
    return join("\n",@l)."\n";
}

sub save()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    push @l,"Final_k_$n = k_$n * fk_$n";
    push @l,"Final_Y_$n = Y_$n * fy_$n";
    push @l,"Final_Fmax_$n = Final_k_$n / 4";
    return join("\n",@l)."\n";
}


sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return cplx(1,0)/($self->value('Y') * ($self->value('k') + cplx(0,1) * 2 * pi() * $f)**0.5);
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fy_$n","fk_$n");
}

sub k_start($$$)
{
    ### Returns a guess at the 'B' parameter of the elemenet ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $x = $l[0];  ### Summit frequency in Hz   ##
    return 4 * $x;
}

sub Y_start($$$)
{
    ### Returns a guess at the 'capacitance' of the element ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $k = 4 * $l[0]; ### gest guess at k parameter                   ##
    @l = split(/[\s\t]+/,$minf);
    my @m = split(/[\s\t]+/,$maxf);
    my $z_max = $l[1] - $m[1];  ### max impedance (intercept with x for element ##
    return 1 / ( $z_max * $k**0.5 );
}

sub helpfunction
{
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return "Select minimum frequency, sumit frequency and maximum frequency for arc" unless (@_);
    return $self->Y_start(@_) if ($tag eq 'Y');
    return $self->k_start(@_) if ($tag eq 'k');
    return undef;
}


package Impedance::Gd;
use Math::Complex;
### Depressed Gerisher impedance ###
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'Gd';
    $self->{'_tags'} = ['Y','k','n'];
    $self->{'_helptags'} = ['Y','k','n'];
    $self->{'_values'}{'n'} = 0.4;
    $self->{'_desc'}{'name'} = "Depressed Gerisher diffusion element";
    $self->{'_desc'}{'Y'} = "'Admittance' of element in S*s^(n)";
    $self->{'_desc'}{'k'} = "'Rate constant' in s^(-1)";
    $self->{'_desc'}{'n'} = "Exponent";
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    return $self->value('k')/(4 * 2.5 * $self->value('n'));
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("Y_$n = ".$self->value('Y'),"fy_$n = 1.0","k_$n = ".$self->value('k'),"n_$n = ".$self->value('n'),"fk_$n = 1.0","fn_$n = 1.0");
    push @l,'z_'.$n.'(x)=1.0 / ( Y_'.$n.' * fy_'.$n.' * (k_'.$n.' * fk_'.$n.' + {0,1}*2*pi*x)**(fn_'.$n.' * n_'.$n.'))';
    return join("\n",@l)."\n";
}

sub save()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    push @l,"Final_k_$n = k_$n * fk_$n";
    push @l,"Final_Y_$n = Y_$n * fy_$n";
    push @l,"Final_n_$n = n_$n * fn_$n";
    push @l,"Final_Fmax_$n = Final_k_$n / 4";
    return join("\n",@l)."\n";
}


sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return cplx(1,0)/($self->value('Y') * ($self->value('k') + cplx(0,1) * 2 * pi() * $f)**$self->value('n'));
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fy_$n","fk_$n","fn_$n");
}

sub n_start($$$)
{
    ### Returns a guess at the 'n' parameter of the elemenet ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$minf);
    my $min = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$maxf);
    my $max = cplx($l[1],$l[2]);
    @l = split(/[\s\t]+/,$sumf);
    my $sum = cplx($l[1],$l[2]);
    my $r = Re($sum) - Re($max);
    my $i = Im($sum) - Im($max);
    return $i / (2*$r);
}

sub k_start($$$)
{
    ### Returns a guess at the 'B' parameter of the elemenet ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $x = $l[0];  ### Summit frequency in Hz   ##
    return 4 * $x;
}

sub Y_start($$$)
{
    ### Returns a guess at the 'capacitance' of the element ###
    my $self = shift;
    my $n = $self->n_start(@_);
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $k = 4 * $l[0]; ### gest guess at k parameter                   ##
    @l = split(/[\s\t]+/,$minf);
    my @m = split(/[\s\t]+/,$maxf);
    my $z_max = $l[1] - $m[1];  ### max impedance (intercept with x for element ##
    return 1 / ( $z_max * $k**$n );
}

sub helpfunction
{
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return "Select minimum frequency, sumit frequency and maximum frequency for arc" unless (@_);
    return $self->Y_start(@_) if ($tag eq 'Y');
    return $self->k_start(@_) if ($tag eq 'k');
    return $self->n_start(@_) if ($tag eq 'n');
    return undef;
}


package Impedance::dL;
use Math::Complex;
### De Levie impedance ###
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'dL';
    $self->{'_tags'} = ['Ri','Rr','Y','n','d'];
#    $self->{'_helptags'} = ['Ri','Rr','Y','n','d'];
    $self->{'_helptags'} = [];
    $self->{'_values'}{'n'} = 0.8;
    $self->{'_values'}{'Ri'} = 10;
    $self->{'_values'}{'Y'} = 1e-2;
    $self->{'_values'}{'d'} = 0.2;
    $self->{'_desc'}{'name'} = "De Levie Impedance";
    $self->{'_desc'}{'Ri'} = "Resistance per unit length of ionic conductor (ohm / cm)";
    $self->{'_desc'}{'Rr'} = "Faradaic reaction resistance unit length (ohm * cm)";
    $self->{'_desc'}{'Y'} = "'Capacitance' of the constant phase element (F*s^(n-1) / cm)";
    $self->{'_desc'}{'n'} = "Phase angle of the constant phase element in '1/2 pi' radians (1 mean 90 degrees)";
    $self->{'_desc'}{'d'} = 'Thickness of porous electrode (cm)';
    bless $self,$class;
    return $self;
}


sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    push @l,'cth(x) = (abs(x)>50 ? {1,0} : cosh(x)/sinh(x))';
    push @l,"Ri_$n =".$self->value('Ri');
    push @l,"fRi_$n = 1.0";
    push @l,"Rr_$n =".$self->value('Rr');
    push @l,"fRr_$n = 1.0";
    push @l,"Y_$n =".$self->value('Y');
    push @l,"fY_$n = 1.0";
    push @l,"d_$n =".$self->value('d');
    push @l,"fd_$n = 1.0";
    push @l,"n_$n =".$self->value('n');
    push @l,"fn_$n = 1.0";
    push @l,'AdL_'.$n.'(x) = (1 + Y_'.$n.' * fY_'.$n.' * ({0,1}*2*pi*x)**(fn_'.$n.' * n_'.$n.'))**0.5';
    push @l,'z_'.$n.'(x)=(Ri_'.$n.'*fRi_'.$n.'*Rr_'.$n.'*fRr_'.$n.')**0.5 * cth( d_'.$n.' * fd_'.$n.' * ((Ri_'.$n.' * fRi_'.$n.' ) / (Rr_'.$n.' * fRr_'.$n.' ))**0.5 * AdL_'.$n.'(x) ) / AdL_'.$n.'(x)';
    return join("\n",@l)."\n";
}


sub save()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    push @l,"Final_n_$n = n_$n * fn_$n";
    push @l,"Final_Ri_$n = Ri_$n * fRi_$n";
    push @l,"Final_Rr_$n = Rr_$n * fRr_$n";
    push @l,"Final_Y_$n = Y_$n * fY_$n";
    push @l,"Final_d_$n = d_$n * fd_$n";
    return join("\n",@l)."\n";
}


sub my_coth($)
{
    my $self = shift;
    my $val = shift;
    if ($val > 50)
    {
	return cplx(1,0);
    }
    else
    {
	return coth($val);
    }
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    my $a = (1 + $self->value('Y') * (cplx(0,1) * 2 * pi() * $f)**$self->value('n'))**0.5;
    my $t = $self->my_coth($self->value('d') * ($self->value('Ri')/$self->value('Rr'))**0.5 * $a);
    return ($self->value('Ri') * $self->value('Rr'))**0.5 * $t / $a; 
}

sub f_imax()
{
    my $self = shift;
    ### Hack based on an RQ, may not be accurate! ###
    return (1/(2 * pi())) * (1/($self->value('Y')*$self->value('R'))) ** (1/$self->value('n'));
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fRi_$n","fRr_$n","fY_$n","fd_$n","fn_$n")
}



return 1;

